#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import random
import socket
import ssl
import string
from queue import Queue, Empty
from re import fullmatch, match, search
from threading import Thread, Event
from time import sleep, time
from typing import List, Union

import orjson
from orjson.orjson import JSONDecodeError

from thingsboard_gateway.connectors.mqtt.backward_compatibility_adapter import BackwardCompatibilityAdapter
from thingsboard_gateway.gateway.constant_enums import Status
from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.connectors.mqtt.mqtt_decorators import CustomCollectStatistics
from thingsboard_gateway.gateway.constants import DATA_RETRIEVING_STARTED, CONVERTED_TS_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.statistics.decorators import CollectAllReceivedBytesStatistics
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_logger import init_logger

try:
    from paho.mqtt.client import Client, Properties
    from paho.mqtt.packettypes import PacketTypes
except ImportError:
    print("paho-mqtt library not found")
    TBUtility.install_package("paho-mqtt")
    from paho.mqtt.client import Client, Properties
    from paho.mqtt.packettypes import PacketTypes

from paho.mqtt.client import MQTTv31, MQTTv311, MQTTv5


MQTT_VERSIONS = {
    3: MQTTv31,
    4: MQTTv311,
    5: MQTTv5
}

RESULT_CODES_V3 = {
    1: "Connection rejected for unsupported protocol version",
    2: "Connection rejected for rejected client ID",
    3: "Connection rejected for unavailable server",
    4: "Connection rejected for damaged username or password",
    5: "Connection rejected for unauthorized",
}

RESULT_CODES_V5 = {
    4:   "Disconnect with Will Message",
    16:  "No matching subscribers",
    17:  "No subscription existed",
    128: "Unspecified error",
    129: "Malformed Packet",
    130: "Protocol Error",
    131: "Implementation specific error",
    132: "Unsupported Protocol Version",
    133: "Client Identifier not valid",
    134: "Bad User Name or Password",
    135: "Not authorized",
    136: "Server unavailable",
    137: "Server busy",
    138: "Banned",
    139: "Server shutting down",
    140: "Bad authentication method",
    141: "Keep Alive timeout",
    142: "Session taken over",
    143: "Topic Filter invalid",
    144: "Topic Name invalid",
    145: "Packet Identifier in use",
    146: "Packet Identifier not found",
    147: "Receive Maximum exceeded",
    148: "Topic Alias invalid",
    149: "Packet too large",
    150: "Message rate too high",
    151: "Quota exceeded",
    152: "Administrative action",
    153: "Payload format invalid",
    154: "Retain not supported",
    155: "QoS not supported",
    156: "Use another server",
    157: "Server moved",
    158: "Shared Subscription not supported",
    159: "Connection rate exceeded",
    160: "Maximum connect time",
    161: "Subscription Identifiers not supported",
    162: "Wildcard Subscription not supported"
}


class MqttConnector(Connector, Thread):
    CONFIGURATION_KEY_SHARED_GLOBAL = "sharedGlobal"
    CONFIGURATION_KEY_SHARED_ID = "sharedId"

    def __init__(self, gateway, config, connector_type):
        super().__init__()

        self.__gateway = gateway  # Reference to TB Gateway
        self._connector_type = connector_type  # Should be "mqtt"

        # Set up lifecycle flags ---------------------------------------------------------------------------------------
        self._connected = False
        self.__stopped = False
        self.__stop_event = Event()
        self.daemon = True

        self.__log = init_logger(self.__gateway, config['name'],
                                 config.get('logLevel', 'INFO'),
                                 enable_remote_logging=config.get('enableRemoteLogging', False),
                                 is_connector_logger=True)
        self.__converter_log = init_logger(self.__gateway, config['name'] + '_converter',
                                           config.get('logLevel', 'INFO'),
                                           enable_remote_logging=config.get('enableRemoteLogging', False),
                                           is_converter_logger=True, attr_name=config['name'])

        # check if the configuration is in the old format
        using_old_config_format_detected = BackwardCompatibilityAdapter.is_old_config_format(config)
        if using_old_config_format_detected:
            self.config = BackwardCompatibilityAdapter(config).convert()
            self.__id = self.config.get('id')
        else:
            self.config = config
            self.__id = self.config.get('id')

        self.statistics = {'MessagesReceived': 0, 'MessagesSent': 0}
        self.__subscribes_sent = {}

        if using_old_config_format_detected:
            self.__log.info("Old MQTT connector configuration format detected. Automatic conversion is applied.")

        # Extract main sections from configuration ---------------------------------------------------------------------
        self.__broker = config.get('broker')
        if not self.__broker:
            self.__log.error('Broker configuration is missing!')
            return

        self.__mapping = []
        self.__server_side_rpc = []
        self.__connect_requests = []
        self.__disconnect_requests = []
        self.__attribute_requests = []
        self.__attribute_updates = []

        self.__shared_custom_converters = {}

        mapping_key = 'mapping' if self.config.get('mapping') else 'dataMapping'

        mandatory_keys = {
            mapping_key: ['topicFilter', 'converter'],
            "serverSideRpc": ['deviceNameFilter', 'methodFilter', 'requestTopicExpression', 'valueExpression'],
            "connectRequests": ['topicFilter'],
            "disconnectRequests": ['topicFilter'],
            "attributeRequests": ['topicFilter', 'topicExpression', 'valueExpression'],
            "attributeUpdates": ['deviceNameFilter', 'attributeFilter', 'topicExpression', 'valueExpression']
        }

        # Mappings, i.e., telemetry/attributes-push handlers provided by user via configuration file
        self.load_handlers(mapping_key, mandatory_keys[mapping_key], self.__mapping)

        # RPCs, i.e., remote procedure calls (ThingsBoard towards devices) handlers
        self.load_handlers('serverSideRpc', mandatory_keys['serverSideRpc'], self.__server_side_rpc)

        # Connect requests, i.e., telling ThingsBoard that a device is online even if it does not post telemetry
        self.load_handlers('connectRequests', mandatory_keys['connectRequests'], self.__connect_requests)

        # Disconnect requests, i.e., telling ThingsBoard that a device is offline even if keep-alive has not expired yet
        self.load_handlers('disconnectRequests', mandatory_keys['disconnectRequests'], self.__disconnect_requests)

        # Shared attributes direct requests, i.e., asking ThingsBoard for some shared attribute value
        self.load_handlers('attributeRequests', mandatory_keys['attributeRequests'], self.__attribute_requests)

        # Attributes updates requests, i.e., asking ThingsBoard to send updates about an attribute
        self.load_handlers('attributeUpdates', mandatory_keys['attributeUpdates'], self.__attribute_updates)

        # Setup topic substitution lists for each class of handlers ----------------------------------------------------
        self.__mapping_sub_topics = {}
        self.__connect_requests_sub_topics = {}
        self.__disconnect_requests_sub_topics = {}
        self.__attribute_requests_sub_topics = {}

        # Set up external MQTT broker connection -----------------------------------------------------------------------
        client_id = self.__broker.get("clientId", ''.join(random.choice(string.ascii_lowercase) for _ in range(23)))

        self._keepAlive = self.__broker.get("keepAlive", 60)
        self._cleanSession = self.__broker.get("cleanSession", True)
        self._cleanStart = self.__broker.get("cleanStart", True)
        self._sessionExpiryInterval = self.__broker.get("sessionExpiryInterval", 0)

        self._mqtt_version = self.__broker.get('version', 5)
        try:
            if self._mqtt_version != 5:
                self._client = Client(client_id=client_id, clean_session=self._cleanSession,
                                      protocol=MQTT_VERSIONS[self._mqtt_version])
            else:
                self._client = Client(client_id=client_id, protocol=MQTT_VERSIONS[self._mqtt_version])
        except KeyError:
            self.__log.error('Unknown MQTT version. Starting up on version 5...')
            self._client = Client(client_id=client_id, protocol=MQTTv5)
            self._mqtt_version = 5

        self.name = config.get("name", self.__broker.get(
            "name",
            'Mqtt Broker ' + ''.join(random.choice(string.ascii_lowercase) for _ in range(5))))

        if "username" in self.__broker["security"]:
            self._client.username_pw_set(self.__broker["security"].get("username"),
                                         self.__broker["security"].get("password"))

        if "caCert" in self.__broker["security"] \
                or self.__broker["security"].get("type", "none").lower() == "certificates":

            self.__log.debug("Connector connecting with certificates")
            ca_cert = self.__broker["security"].get("pathToCACert")
            private_key = self.__broker["security"].get("pathToPrivateKey")
            cert = self.__broker["security"].get("pathToClientCert")
            if ca_cert is None:
                self._client.tls_set_context(ssl.SSLContext(ssl.PROTOCOL_TLSv1_2))
            else:
                try:
                    self._client.tls_set(ca_certs=ca_cert,
                                         certfile=cert,
                                         keyfile=private_key,
                                         cert_reqs=ssl.CERT_REQUIRED,
                                         tls_version=ssl.PROTOCOL_TLSv1_2,
                                         ciphers=None)
                except Exception as e:
                    self.__log.error("Cannot setup connection to broker %s using SSL. "
                                     "Please check your configuration.\nError: %s",
                                     self.get_name(), e)
                if self.__broker["security"].get("insecure", False):
                    self._client.tls_insecure_set(True)
                    self.__log.debug("Connector tls_insecure_set: True")
                else:
                    self._client.tls_insecure_set(False)
                    self.__log.debug("Connector tls_insecure_set: False")
        else:
            self.__log.debug("Connector connecting anonymously")
        # Set up external MQTT broker callbacks ------------------------------------------------------------------------
        self._client.on_connect = self._on_connect
        self._client.on_message = self._on_message
        self._client.on_subscribe = self._on_subscribe
        self._client.on_disconnect = self._on_disconnect
        # self._client.on_log = self._on_log

        self.__msg_queue = Queue(self.__broker.get('maxMessageQueue', 1000000000))
        self.__workers_thread_pool = []
        self.__max_msg_number_for_worker = self.__broker.get('maxMessageNumberPerWorker', 10)
        self.__max_number_of_workers = self.__broker.get('maxNumberOfWorkers', 100)

        self._on_message_queue = Queue(self.__broker.get('maxProcessingMessageQueue', 1000000000))
        self._on_message_thread = Thread(name='On Message', target=self._process_on_message, daemon=True)
        self._on_message_thread.start()

    def get_config(self):
        return self.config

    def get_type(self):
        return self._connector_type

    @staticmethod
    def __add_ts_to_test_message(msg, ts_name):
        msg = orjson.loads(msg.decode('utf-8').replace("'", '"'))
        msg[ts_name] = int(time() * 1000)
        return orjson.dumps(msg).encode('utf-8')

    def load_handlers(self, handler_flavor, mandatory_keys, accepted_handlers_list):
        handler_configuration = self.config.get(handler_flavor)
        if handler_configuration is None:
            request_mapping_config = self.config.get("requestsMapping", {})
            if isinstance(request_mapping_config, dict):
                handler_configuration = None
                for request_mapping in request_mapping_config:
                    if request_mapping == handler_flavor:
                        handler_configuration = request_mapping_config[request_mapping]

        if not handler_configuration:
            self.__log.debug("'%s' section missing from configuration", handler_flavor)
        else:
            for handler in handler_configuration:
                discard = False

                for key in mandatory_keys:
                    if key not in handler:
                        # Will report all missing fields to user before discarding the entry => no break here
                        discard = True
                        self.__log.error("Mandatory key '%s' missing from %s handler: %r",
                                         key, handler_flavor, handler)
                    else:
                        self.__log.debug("Mandatory key '%s' found in %s handler: %r",
                                         key, handler_flavor, handler)

                if discard:
                    self.__log.warning("%s handler is missing some mandatory keys => rejected: %r",
                                       handler_flavor, handler)
                else:
                    accepted_handlers_list.append(handler)
                    self.__log.debug("%s handler has all mandatory keys => accepted: %r",
                                     handler_flavor, handler)

            self.__log.info("Number of accepted %s handlers: %d",
                            handler_flavor,
                            len(accepted_handlers_list))

            self.__log.debug("Number of rejected %s handlers: %d",
                             handler_flavor,
                             len(handler_configuration) - len(accepted_handlers_list))

    def is_connected(self):
        return self._connected

    def is_stopped(self):
        return self.__stopped

    def open(self):
        self.__stopped = False
        self.start()

    def run(self):
        try:
            self.__connect()
        except Exception as e:
            self.__log.exception("Error in connector main loop: %s", e)

        while not self.__stopped:
            try:
                if not self._connected:
                    self.__connect()

                self.__threads_manager()

                self.__stop_event.wait(timeout=0.2)
            except TimeoutError:
                pass
            except Exception as e:
                self.__log.exception("Error in connector main loop: %s", e)

    def __connect(self):
        while not self._connected and not self.__stopped:
            try:
                if self._mqtt_version != 5:
                    self._client.connect(self.__broker['host'],
                                         self.__broker.get('port', 1883),
                                         keepalive=self._keepAlive)
                else:
                    properties = Properties(PacketTypes.CONNECT)
                    properties.SessionExpiryInterval = self._sessionExpiryInterval
                    self._client.connect(self.__broker['host'],
                                         self.__broker.get('port', 1883),
                                         keepalive=self._keepAlive,
                                         clean_start=self._cleanStart,
                                         properties=properties)
                self._client.loop_start()
                if not self._connected:
                    sleep(1)
            except (ConnectionRefusedError, ConnectionResetError, ssl.SSLEOFError, socket.timeout) as e:
                self.__log.error("Error while connecting to broker %s: %s", self.get_name(), e)
                sleep(10)
            except Exception as e:
                self.__log.exception("Error while connecting to broker %s: %s", self.get_name(), e)
                sleep(10)

    def close(self):
        self.__stopped = True
        self.__stop_event.set()
        try:
            self._client.disconnect()
        except Exception as e:
            self.__log.exception(e)
        self._client.loop_stop()
        for worker in self.__workers_thread_pool:
            worker.stop()
        self.__log.info('%s has been stopped.', self.get_name())
        self.__log.stop()

    def get_name(self):
        return self.name

    def get_id(self):
        return self.__id

    def __subscribe(self, topic, qos):
        message = self._client.subscribe(topic, qos)
        try:
            self.__subscribes_sent[message[1]] = topic
        except Exception as e:
            self.__log.exception(e)

    def _on_connect(self, client, userdata, flags, result_code, *extra_params):
        if result_code == 0:
            self._connected = True
            self.__log.info('%s connected to %s:%s - successfully.',
                            self.get_name(),
                            self.__broker["host"],
                            self.__broker.get("port", "1883"))

            self.__log.debug("Client %s, userdata %s, flags %s, extra_params %s",
                             str(client),
                             str(userdata),
                             str(flags),
                             extra_params)

            self.__mapping_sub_topics = {}

            # Setup data upload requests handling ----------------------------------------------------------------------
            for mapping in self.__mapping:
                try:
                    # Load converter for this mapping entry ------------------------------------------------------------
                    # mappings are guaranteed to have topicFilter and converter fields. See __init__
                    default_converters = {
                        "json": "JsonMqttUplinkConverter",
                        "bytes": "BytesMqttUplinkConverter"
                    }

                    # Get converter class from "extension" parameter or default converter
                    converter_class_name = mapping["converter"].get("extension",
                                                                    default_converters.get(
                                                                        mapping['converter'].get('type')))
                    if not converter_class_name:
                        self.__log.error('Converter type or extension class should be configured!')
                        continue

                    converter = None
                    sharing_id = None
                    if mapping["converter"].get("type") == "custom" or mapping["converter"].get("extension"):
                        if mapping["converter"].get(self.CONFIGURATION_KEY_SHARED_GLOBAL, False):
                            sharing_id = converter_class_name
                        elif mapping["converter"].get(self.CONFIGURATION_KEY_SHARED_ID):
                            sharing_id = mapping["converter"][self.CONFIGURATION_KEY_SHARED_ID]

                        if sharing_id:
                            converter = self.__shared_custom_converters.get(sharing_id)
                            self.__log.debug('Converter %s for topic %s will use in sharing mode!', sharing_id,
                                             mapping["topicFilter"])

                    if not converter:
                        # Find and load required class
                        module = TBModuleLoader.import_module(self._connector_type, converter_class_name)
                        if module:
                            self.__log.debug('Converter %s for topic %s - found!', converter_class_name,
                                             mapping["topicFilter"])
                            converter = module(mapping, self.__converter_log)
                            if sharing_id:
                                self.__shared_custom_converters[sharing_id] = converter
                        else:
                            self.__log.error("Cannot find converter for %s topic", mapping["topicFilter"])
                            continue
                    else:
                        self.__log.debug('Converter %s for topic %s - found in cache!', converter_class_name,
                                         mapping["topicFilter"])

                    # Setup regexp topic acceptance list ---------------------------------------------------------------
                    # Check if topic is shared subscription type
                    # (an exception is aws topics that do not support shared subscription)
                    regex_topic = mapping["topicFilter"]
                    if not regex_topic.startswith('$aws') and regex_topic.startswith('$'):
                        regex_topic = '/'.join(regex_topic.split('/')[2:])
                    else:
                        regex_topic = TBUtility.topic_to_regex(regex_topic)

                    # There may be more than one converter per topic, so I'm using vectors
                    if not self.__mapping_sub_topics.get(regex_topic):
                        self.__mapping_sub_topics[regex_topic] = []

                    self.__mapping_sub_topics[regex_topic].append(converter)

                    # Subscribe to appropriate topic -------------------------------------------------------------------
                    self.__subscribe(mapping["topicFilter"], mapping.get("subscriptionQos", 1))

                    self.__log.info('Connector "%s" subscribe to %s',
                                    self.get_name(),
                                    TBUtility.regex_to_topic(regex_topic))

                except Exception as e:
                    self.__log.exception(e)

            self.__setup_request_subscriptions(self.__connect_requests, self.__connect_requests_sub_topics)
            self.__setup_request_subscriptions(self.__disconnect_requests, self.__disconnect_requests_sub_topics)
            self.__setup_request_subscriptions(self.__attribute_requests, self.__attribute_requests_sub_topics)
        else:
            result_codes = RESULT_CODES_V5 if self._mqtt_version == 5 else RESULT_CODES_V3
            rc = result_code.value if self._mqtt_version == 5 else result_code
            if rc in result_codes:
                self.__log.error("%s connection FAIL with error %s %s!", self.get_name(), rc, result_codes[rc])
            else:
                self.__log.error("%s connection FAIL with unknown error!", self.get_name())

    def __setup_request_subscriptions(self, requests: list, target_dict: dict):
        for request in [entry for entry in requests if entry is not None]:
            try:
                self.__subscribe(request["topicFilter"], request.get("subscriptionQos", 1))
                topic_filter = TBUtility.topic_to_regex(request.get("topicFilter"))
                target_dict[topic_filter] = request

            except KeyError as e:
                self.__log.error("Failed to extract required parts of request to topic %s", str(e))
                self.__log.debug("Error", exc_info=True)
                continue

    def _on_disconnect(self, *args):
        self._connected = False
        self.__log.debug('"%s" was disconnected. %s', self.get_name(), str(args))

    def _on_log(self, *args):
        self.__log.debug(args)

    def _on_subscribe(self, _, __, mid, granted_qos, *args):
        self.__log.info(args)
        try:
            if granted_qos[0] == 128:
                self.__log.error('"%s" subscription failed to topic %s subscription message id = %i',
                                 self.get_name(),
                                 self.__subscribes_sent.get(mid), mid)
            else:
                self.__log.info('"%s" subscription success to topic %s, subscription message id = %i',
                                self.get_name(),
                                self.__subscribes_sent.get(mid), mid)
        except Exception as e:
            self.__log.exception(e)

        # Success or not, remove this topic from the list of pending subscription requests
        if self.__subscribes_sent.get(mid) is not None:
            del self.__subscribes_sent[mid]

    def put_data_to_convert(self, converter, message, content) -> bool:
        if not self.__msg_queue.full():
            if not hasattr(converter, 'SUPPORTS_BYTES_PAYLOAD'):
                content = TBUtility.decode(content)
            self.__msg_queue.put((converter.convert, message.topic, content), True, 100)
            return True
        return False

    def _save_converted_msg(self, topic, data):
        data.add_to_metadata({DATA_RETRIEVING_STARTED: int(time() * 1000)})
        if self.__gateway.send_to_storage(self.name, self.get_id(), data) == Status.SUCCESS:
            StatisticsService.count_connector_message(self.name, stat_parameter_name='storageMsgPushed')
            self.statistics['MessagesSent'] += 1
            self.__log.debug("Successfully converted message from topic %s", topic)

    def __threads_manager(self):
        if len(self.__workers_thread_pool) == 0:
            worker = MqttConnector.ConverterWorker("Main Worker", self.__msg_queue, self._save_converted_msg)
            self.__workers_thread_pool.append(worker)
            worker.start()

        number_of_needed_threads = round(self.__msg_queue.qsize() / self.__max_msg_number_for_worker, 0)
        threads_count = len(self.__workers_thread_pool)
        if number_of_needed_threads > threads_count < self.__max_number_of_workers:
            thread = MqttConnector.ConverterWorker(
                "Worker " + ''.join(random.choice(string.ascii_lowercase) for _ in range(5)), self.__msg_queue,
                self._save_converted_msg)
            self.__workers_thread_pool.append(thread)
            thread.start()
        elif number_of_needed_threads < threads_count and threads_count > 1:
            worker: MqttConnector.ConverterWorker = self.__workers_thread_pool[-1]
            worker.stopped = True
            self.__workers_thread_pool.remove(worker)

    def _on_message(self, client, userdata, message):
        StatisticsService.count_connector_message(self.name, stat_parameter_name='connectorMsgsReceived')
        StatisticsService.count_connector_bytes(self.name, message.payload,
                                                stat_parameter_name='connectorBytesReceived')
        self._on_message_queue.put((client, userdata, message))

    def _parse_device_info(self, device_info, topic, content):
        found_device_name = None
        found_device_type = 'default'

        # Get device name, either from topic or from content
        try:
            if device_info.get('deviceNameExpressionSource') == 'topic':
                device_name_match = search(device_info["deviceNameExpression"], topic)
                if device_name_match is not None:
                    found_device_name = device_name_match.group(0)
            elif device_info.get('deviceNameExpressionSource') == 'message':
                found_device_name = TBUtility.get_value(device_info["deviceNameExpression"], content,
                                                        expression_instead_none=True)
            elif device_info.get('deviceNameExpressionSource') == 'constant':
                found_device_name = device_info["deviceNameExpression"]

            # Get device type (if any), either from topic or from content
            if device_info.get("deviceProfileExpressionSource") == 'topic':
                device_type_match = search(device_info["deviceProfileExpression"], topic)
                found_device_type = device_type_match.group(0) if device_type_match is not None else device_info[
                    "deviceProfileExpression"]
            elif device_info.get("deviceProfileExpressionSource") == 'message':
                found_device_type = TBUtility.get_value(device_info["deviceProfileExpression"], content,
                                                        expression_instead_none=True)
            elif device_info.get("deviceProfileExpressionSource") == 'constant':
                found_device_type = device_info["deviceProfileExpression"]
            return found_device_name, found_device_type

        except JSONDecodeError as e:
            self.__log.error("Check your message payload for an incorrect format: %s", str(e))
            self.__log.debug("Error %s", e, exc_info=True)
            return None, None

        except Exception as e:
           self.__log.exception("An unexpected error occurred while parsing device info: %s", str(e))
           self.__log.debug("Error %s", e, exc_info=True)
           return None, None

    def _process_on_message(self):
        while not self.__stopped:
            if not self._on_message_queue.empty():
                client, userdata, message = self._on_message_queue.get_nowait()

                self.statistics['MessagesReceived'] += 1
                content = None

                # Check if message topic exists in mappings "i.e., I'm posting telemetry/attributes" -------------------
                topic_handlers = [regex for regex in self.__mapping_sub_topics if fullmatch(regex, message.topic)]

                if topic_handlers:
                    # Note: every topic may be associated to one or more converter.
                    # This means that a single MQTT message
                    # may produce more than one message towards ThingsBoard. This also means that I cannot return after
                    # the first successful conversion: I got to use all the available ones.
                    # I will use a flag to understand whether at least one converter succeeded
                    request_handled = False

                    for topic in topic_handlers:
                        available_converters = self.__mapping_sub_topics[topic]
                        for converter in available_converters:
                            try:
                                request_handled = self.put_data_to_convert(converter, message, message.payload)
                            except Exception as e:
                                self.__log.exception(e)

                    if not request_handled:
                        self.__log.error('Cannot find converter for the topic:"%s"! Client: %s, User data: %s',
                                         message.topic,
                                         str(client),
                                         str(userdata))

                    # Note: if I'm in this branch, this was for sure a telemetry/attribute push message
                    # => Execution must end here both in case of failure and success
                    continue

                # The main request processing block, the try/except statements are added to avoid whole attributes processing
                # to be stopped because of a single error in a request processing
                try:

                    # Handling connect requests ----------------------------------------------------------------
                    request_handled, content = self.__process_connect(message, content)
                    if request_handled:
                        continue

                    # Handling disconnect requests ----------------------------------------------------------------
                    request_handled, content = self.__process_disconnect(message, content)
                    if request_handled:
                        continue

                    # Handling attribute requests ----------------------------------------------------------------
                    request_handled, content = self.__process_attribute_request(message, content)
                    if request_handled:
                        continue

                # In case of failure in any block above, log the error and continue
                except TypeError as e:
                    self.__log.exception("Make sure your input match with config and the payload you sent was valid.",)
                    continue

                except Exception as e:
                    self.__log.exception("An unexpected error occurred while processing request: %s", str(e))
                    self.__log.debug("Error", exc_info=True)
                    continue

                # Check if message topic exists in RPC handlers --------------------------------------------------------
                # The gateway is expecting for this message => no wildcards here, the topic must be evaluated as is

                if self.__gateway.is_rpc_in_progress(message.topic):
                    content = message.payload.decode('utf-8').replace("'", '"')
                    self.__log.info("RPC response arrived. Forwarding it to thingsboard.")
                    self.__gateway.rpc_with_reply_processing(message.topic, content)
                    continue

                self.__log.debug("Received message to topic \"%s\" with unknown interpreter data: \n\n\"%s\"",
                                 message.topic,
                                 content)
            else:
                sleep(.2)

    def __process_connect(self, message, content):
        topic_handlers = self.__match_handlers(self.__connect_requests_sub_topics, message.topic)
        if not topic_handlers:
            return False, content
        content = self.__decode_content_from_message(message, content)

        for regex in topic_handlers:
            handler = self.__connect_requests_sub_topics[regex]
            found_device_name, found_device_type = self.__resolve_device_name(handler, message.topic, content)

            if found_device_name is None:
                self.__log.error("Device name missing from connection request")
                continue

            self.__log.info("Connecting device %s of type %s", found_device_name, found_device_type)
            self.__gateway.add_device(found_device_name, {"connector": self}, device_type=found_device_type)

        return True, content

    def __process_disconnect(self, message, content):
        topic_handlers = self.__match_handlers(self.__disconnect_requests_sub_topics, message.topic)
        if not topic_handlers:
            return False, content
        content = self.__decode_content_from_message(message, content)

        for regex in topic_handlers:
            handler = self.__disconnect_requests_sub_topics[regex]
            found_device_name, found_device_type = self.__resolve_device_name(handler, message.topic, content)

            if found_device_name is None:
                self.__log.error("Device name missing from disconnection request")
                continue

            if found_device_name in self.__gateway.get_devices():
                self.__log.info("Disconnecting device %s of type %s", found_device_name, found_device_type)
                self.__gateway.del_device(found_device_name)

            else:
                self.__log.info("Device %s was not connected", found_device_name)

            break

        return True, content

    def __process_attribute_request(self, message, content):
        topic_handlers = self.__match_handlers(self.__attribute_requests_sub_topics, message.topic)
        if not topic_handlers:
            return False, content
        content = self.__decode_content_from_message(message, content)

        try:
            for regex in topic_handlers:
                handler = self.__attribute_requests_sub_topics[regex]
                found_device_name, _ = self.__resolve_device_name(handler, message.topic, content)

                if found_device_name is None:
                    self.__log.error("Device name missing from attribute request")
                    continue
                found_attribute_names = self.__extract_requested_arguments_names(handler, message.topic, content)

                if not found_attribute_names:
                    self.__log.error("Attribute name missing from attribute request")
                    continue

                self.__log.debug("Retrieved attribute names %s for device %s", found_attribute_names)

                scope = handler.get('scope') or 'shared'
                formated_value = TBUtility.get_value(f'${scope}', content, get_tag=False)
                if content and formated_value is not None:
                    scope = formated_value
                request_arguments = (found_device_name, found_attribute_names,
                                     lambda data, *args: self.notify_attribute(data, found_attribute_names,
                                                                               handler.get("topicExpression"),
                                                                               handler.get("valueExpression"),
                                                                               handler.get('retain', False),
                                                                               handler.get('qos', 0)))
                if scope == 'client':
                    self.__gateway.tb_client.client.gw_request_client_attributes(*request_arguments)
                    self.__log.info("Successfully processed client attribute request for %s of %s",
                                    found_attribute_names, found_device_name)
                else:
                    self.__gateway.tb_client.client.gw_request_shared_attributes(*request_arguments)
                    self.__log.info("Successfully processed shared attribute request for %s of %s",
                                    found_attribute_names, found_device_name)
                break
            return True, content

        except Exception as e:
            self.__log.exception("An unexpected error occurred while processing attribute request: %s", str(e))
            self.__log.debug("Exception: %s", e, exc_info=True)

    def notify_attribute(self, incoming_data, attribute_name, topic_expression, value_expression, retain, qos):
        if incoming_data.get("device") is None or incoming_data.get("value", incoming_data.get('values')) is None:
            return

        if len(attribute_name) == 0:
            self.__log.error("Attribute name is empty, cannot send attribute data to device")
            return

        device_name = incoming_data.get("device")
        attribute_values = incoming_data.get("value", incoming_data.get("values"))
        try:

            topic = topic_expression \
                .replace("${deviceName}", str(device_name)) \
                .replace("${attributeKey}", str(attribute_name))

            if len(attribute_name) <= 1:
                data = value_expression.replace("${attributeKey}", str(attribute_name[0])) \
                    .replace("${attributeValue}", str(attribute_values))
            else:
                data = orjson.dumps(attribute_values)
                self.__log.debug("Attribute data: %s for device %s to topic: %s", data, device_name, topic)

        except KeyError as e:
            self.__log.error("Cannot form topic/value for attribute request %s due to", incoming_data, str(e))
            return

        except Exception as e:
            self.__log.error(
                "An unexpected error occurred while forming attribute message %s for attribute request:", str(e),
                incoming_data)
            self.__log.debug("Exception: %s", e, exc_info=True)
            return

        self._client.publish(topic, data, qos=qos, retain=retain).wait_for_publish()

    @staticmethod
    def __match_handlers(sub_topics_dict, topic):
        return [regex for regex in sub_topics_dict if fullmatch(regex, topic)]

    @staticmethod
    def __decode_content_from_message(message, content):
        return TBUtility.decode(message) if content is None else content

    @staticmethod
    def __extract_requested_arguments_names(handler, topic, content):
        attribute_name_expression_source = handler.get("attributeNameExpressionSource")

        if attribute_name_expression_source == "topic":
            attribute_name_match = search(attribute_name_expression_source, topic)
            return attribute_name_match.group(0) if attribute_name_match is not None else None

        elif attribute_name_expression_source in ("message", "constant"):
            requested_shared_attributes = TBUtility.get_values(handler.get("attributeNameExpression", ""), content)
            return list(filter(lambda x: x is not None, requested_shared_attributes))

        return None

    def __resolve_device_name(self, handler, topic, content):
        device_info = handler.get("deviceInfo", {})
        found_device_name, found_device_type = self._parse_device_info(device_info, topic, content)
        return found_device_name, found_device_type

    @staticmethod
    def _format_value(value):
        if isinstance(value, (dict, list)):
            formatted_value = orjson.dumps(value).decode('utf-8')
        elif isinstance(value, bool):
            formatted_value = str(value).lower()
        elif value is None:
            formatted_value = "null"
        else:
            formatted_value = str(value)
        return formatted_value

    @CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def on_attributes_update(self, content):
        self.__log.debug('Got attributes update: %s', content)
        try:
            if not self.__attribute_updates:
                self.__log.error('Attributes update config not found.')
                return
            device_content = content['device']
            data = content['data']
            filtered_attribute_update_configuration = [attribute_update_section for attribute_update_section in
                                                       self.__attribute_updates if
                                                       match(attribute_update_section["deviceNameFilter"],
                                                             device_content)]
            if not filtered_attribute_update_configuration:
                self.__log.error("Cannot find deviceName by filter in message with data: %s", content)
                return

            for attribute_update_configuration in filtered_attribute_update_configuration:
                topic_expression = attribute_update_configuration.get("topicExpression")
                value_expression = attribute_update_configuration.get("valueExpression")
                attribute_filter = attribute_update_configuration.get("attributeFilter")

                if not topic_expression or not value_expression or not attribute_filter:
                    self.__log.error(
                        "Cannot find topicExpression or valueExpression or attributeFilter in attribute update configuration for: %s",
                        attribute_update_configuration)
                    continue

                for attribute_key, attribute_value in data.items():
                    if not match(attribute_filter, attribute_key):
                        self.__log.error("Attribute %s does not match filter %s, skipping", attribute_key,
                                         attribute_filter)
                        continue

                    formated_value = self._format_value(attribute_value)
                    try:
                        topic = topic_expression \
                            .replace("${deviceName}", str(device_content)) \
                            .replace("${attributeKey}", str(attribute_key)) \
                            .replace("${attributeValue}", formated_value)
                        data_to_send = value_expression \
                            .replace("${attributeKey}", str(attribute_key)) \
                            .replace("${attributeValue}", formated_value) \
                            .replace("${deviceName}", str(device_content))
                        self._publish(topic, data_to_send,
                                      attribute_update_configuration.get('retain', False),
                                      attribute_update_configuration.get('qos', 0))
                        self.__log.debug("Attribute Update data: %s for device %s to topic: %s", data_to_send,
                                         device_content, topic)

                    except KeyError as e:
                        self.__log.exception("Cannot form topic/value for attribute configuration section %s",
                                             attribute_update_configuration)
        except Exception as e:
            self.__log.exception('Error while processing attributes update %s', str(e))
            self.__log.debug("Exception: %s", e, exc_info=True)

    def __process_rpc_request(self, content, rpc_config):
        try:
            # This handler seems able to handle the request
            self.__log.info("Candidate RPC handler found")

            expects_response = rpc_config.get("responseTopicExpression")
            defines_timeout = rpc_config.get("responseTimeout")

            # 2-way RPC setup
            if expects_response and defines_timeout:
                expected_response_topic = rpc_config["responseTopicExpression"] \
                    .replace("${methodName}", str(content['data']['method'])) \
                    .replace("${requestId}", str(content["data"]["id"]))

                if content.get('device'):
                    expected_response_topic = expected_response_topic.replace("${deviceName}", str(content["device"]))

                expected_response_topic = TBUtility.replace_params_tags(expected_response_topic, content)

                timeout = time() * 1000 + rpc_config.get("responseTimeout")

                # Start listening on the response topic
                self.__log.info("Subscribing to: %s", expected_response_topic)
                self.__subscribe(expected_response_topic, rpc_config.get("responseTopicQoS", 1))

                # Wait for subscription to be carried out
                sub_response_timeout = 10

                while expected_response_topic in self.__subscribes_sent.values():
                    sub_response_timeout -= 1
                    sleep(0.1)
                    if sub_response_timeout == 0:
                        break

                # Ask the gateway to enqueue this as an RPC response
                self.__gateway.register_rpc_request_timeout(content,
                                                            timeout,
                                                            expected_response_topic,
                                                            self.rpc_cancel_processing)

                # Wait for RPC to be successfully enqueued, which never fails.
                while self.__gateway.is_rpc_in_progress(expected_response_topic):
                    sleep(0.1)

            elif expects_response and not defines_timeout:
                self.__log.info("2-way RPC without timeout: treating as 1-way")

            # Actually reach out for the device
            request_topic: str = rpc_config.get("requestTopicExpression") \
                .replace("${methodName}", str(content['data']['method'])) \
                .replace("${requestId}", str(content["data"]["id"]))

            if content.get('device'):
                request_topic = request_topic.replace("${deviceName}", str(content["device"]))

            request_topic = TBUtility.replace_params_tags(request_topic, content)

            data_to_send_tags = TBUtility.get_values(rpc_config.get('valueExpression'), content['data'],
                                                     'params',
                                                     get_tag=True)
            data_to_send_values = TBUtility.get_values(rpc_config.get('valueExpression'), content['data'],
                                                       'params',
                                                       expression_instead_none=True)

            data_to_send = rpc_config.get('valueExpression')

            if content.get('device'):
                data_to_send = data_to_send.replace("${deviceName}", str(content["device"]))

            for (tag, value) in zip(data_to_send_tags, data_to_send_values):
                data_to_send = data_to_send.replace('${' + tag + '}', orjson.dumps(value).decode('utf-8'))

            try:
                self.__log.info("Publishing to: %s with data %s", request_topic, data_to_send)
                result = None
                try:
                    result = self._publish(request_topic, data_to_send, rpc_config.get('retain', False), rpc_config.get('qos', 0))
                except Exception as e:
                    self.__log.exception("Error during publishing to target broker: %r", e)
                    self.__gateway.send_rpc_reply(device=content.get("device"),
                                                  req_id=content["data"]["id"],
                                                  content={
                                                      "error": str.format("Error on publish to target broker: %r",
                                                                          str(e))},
                                                  success_sent=False, to_connector_rpc=True if content.get('device') is None else False) # noqa
                    return
                if not expects_response or not defines_timeout:
                    self.__log.info("One-way RPC: sending ack to ThingsBoard immediately")
                    self.__gateway.send_rpc_reply(device=content.get('device'), req_id=content["data"]["id"],
                                                  success_sent=result is not None, to_connector_rpc=True if content.get('device') is None else False) # noqa

                # Everything went out smoothly: RPC is served
                return
            except Exception as e:
                self.__log.exception("Error during publishing RPC response: ", exc_info=e)
        except Exception as e:
            self.__log.exception("Error during processing RPC request: ", exc_info=e)

    @CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def server_side_rpc_handler(self, content):
        try:
            self.__log.info("Incoming server-side RPC: %s", content)

            if content.get('data') is None:
                content['data'] = {'params': content['params'], 'method': content['method'], 'id': content['id']}

            rpc_method = content['data']['method']

            # check if RPC type is connector RPC (can be only 'get' or 'set')
            try:
                (connector_type, rpc_method_name) = rpc_method.split('_')
                if connector_type == self._connector_type:
                    rpc_method = rpc_method_name
            except ValueError:
                pass

            if content.get('device'):
                # check if RPC method is reserved get/set
                if rpc_method == 'get' or rpc_method == 'set':
                    params = {}
                    for param in content['data']['params'].split(';'):
                        try:
                            (key, value) = param.split('=')
                        except ValueError:
                            continue

                        if key and value:
                            params[key] = value

                    params['valueExpression'] = params.pop('value', None)

                    return self.__process_rpc_request(content, params)
                else:
                    # Check whether one of my RPC handlers can handle this request
                    for rpc_config in self.__server_side_rpc:
                        if search(rpc_config["deviceNameFilter"], content["device"]) \
                                and search(rpc_config["methodFilter"], rpc_method) is not None:
                            return self.__process_rpc_request(content, rpc_config)

                    self.__log.error("RPC not handled: %s", content)
            else:
                return self.__process_rpc_request(content, content['data']['params'])
        except Exception as e:
            self.__log.exception("Error during handling RPC request", exc_info=e)

    @CustomCollectStatistics(start_stat_type='allBytesSentToDevices')
    def _publish(self, request_topic, data_to_send, retain, qos):
        result = False
        try:
            if self._connected and self._client is not None and self._client.is_connected():
                self._client.publish(request_topic, data_to_send, qos=qos, retain=retain).wait_for_publish()
                result = True
        except Exception as e:
            self.__log.error("Error during publishing to target broker: %r", e)
        return result

    def rpc_cancel_processing(self, topic):
        self.__log.info("RPC canceled or terminated. Unsubscribing from %s", topic)
        self._client.unsubscribe(topic)

    def get_converters(self):
        return [item[0] for _, item in self.__mapping_sub_topics.items()]

    def _send_current_converter_config(self, name, config):
        self.__gateway.send_attributes({name: config})

    class ConverterWorker(Thread):
        def __init__(self, name, incoming_queue, send_result, batch_size=100):
            super().__init__()
            self.stopped = False
            self.name = name
            self.daemon = True
            self.__msg_queue = incoming_queue
            self.__send_result = send_result
            self.__batch_size = batch_size
            self.__stop_event = Event()

        def run(self):
            while not self.stopped:
                try:
                    batch = []
                    for _ in range(self.__batch_size):
                        try:
                            batch.append(self.__msg_queue.get_nowait())
                        except Empty:
                            break

                    if not batch:
                        self.__stop_event.wait(0.01)
                        continue

                    for convert_function, config, incoming_data in batch:
                        converted_data: Union[ConvertedData, List[ConvertedData]] = convert_function(config, incoming_data)
                        if isinstance(converted_data, ConvertedData):
                            converted_data.add_to_metadata({CONVERTED_TS_PARAMETER: int(time() * 1000)})
                            if converted_data and (converted_data.telemetry_datapoints_count > 0 or
                                                   converted_data.attributes_datapoints_count > 0):
                                self.__send_result(config, converted_data)
                        else:
                            for data in converted_data:
                                data.add_to_metadata({CONVERTED_TS_PARAMETER: int(time() * 1000)})
                                if data.telemetry_datapoints_count > 0 or data.attributes_datapoints_count > 0:
                                    self.__send_result(config, data)
                except Exception as e:
                    # Log the exception if needed
                    print("Error in worker: ", e)
                    pass

        def stop(self):
            self.stopped = True
            self.__stop_event.set()



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from copy import deepcopy


class BackwardCompatibilityAdapter:
    def __init__(self, config, logger):
        self.__log = logger
        self.__config = deepcopy(config)

    def convert(self):
        try:
            self.__convert_application_section()
            self.__convert_device_section()
        except Exception as e:
            self.__log.error('Error converting old config: %s', e)

        return self.__config

    def __convert_application_section(self):
        general_section = self.__config.pop('general', {})
        self.__convert_address_in_section(general_section)
        self.__config['application'] = general_section

    def __convert_device_section(self):
        for device in self.__config.get('devices', []):
            self.__convert_device_info_section(device)
            self.__convert_address_in_section(device)

            for section in ('attributes', 'timeseries'):
                for item_config in device.get(section, []):
                    self.__convert_object_id(item_config)

            for attr_update_config in device.get('attributeUpdates', []):
                self.__convert_object_id(attr_update_config)

            for rpc_config in device.get('rpc', []):
                self.__convert_object_id(rpc_config)

    @staticmethod
    def __convert_address_in_section(old_config_section):
        address = old_config_section.get('address', '')
        if address:
            return old_config_section
        if 'host' in old_config_section:
            address += old_config_section['host']
            address = address.rstrip('/')
        if 'mask' in old_config_section:
            network_mask = old_config_section['mask']
            if network_mask:
                address += '/' + (str(network_mask) if not isinstance(network_mask, str) else network_mask)
        if 'port' in old_config_section:
            address += ':' + (str(old_config_section['port']) if not isinstance(old_config_section['port'], str) else old_config_section['port'])  # noqa
        old_config_section['address'] = address

    @staticmethod
    def __convert_device_info_section(old_device_config):
        device_name = old_device_config.pop('deviceName')
        device_type = old_device_config.pop('deviceType', 'default')

        old_device_config['deviceInfo'] = {
            'deviceNameExpressionSource': 'expression',
            'deviceProfileExpressionSource': 'expression',
            'deviceNameExpression': device_name,
            'deviceProfileExpression': device_type
        }

    @staticmethod
    def __convert_object_id(old_item_config):
        old_object_id = old_item_config.pop('objectId')
        (object_type, object_id) = old_object_id.split(':')
        old_item_config['objectType'] = object_type
        old_item_config['objectId'] = object_id

    @staticmethod
    def is_old_config(config):
        return config.get('general') is not None



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import random
import socket
import ssl
import string
import threading
import inspect
from logging import getLogger
from os import environ
from os.path import exists, join, abspath, dirname
from sys import path
from ssl import CERT_REQUIRED
from copy import deepcopy
from time import sleep, time
from typing import Union
import socks

from simplejson import dumps, load

from thingsboard_gateway.gateway.constants import DEV_MODE_PARAMETER_NAME, PROVISIONED_CREDENTIALS_FILENAME
from thingsboard_gateway.tb_utility.tb_utility import TBUtility

try:
    if environ.get(DEV_MODE_PARAMETER_NAME) is not None and environ.get(DEV_MODE_PARAMETER_NAME).lower() == 'true':
        raise ImportError
    from tb_gateway_mqtt import TBGatewayMqttClient, TBDeviceMqttClient, \
        GATEWAY_ATTRIBUTES_RESPONSE_TOPIC, GATEWAY_ATTRIBUTES_TOPIC
    import tb_device_mqtt
except ImportError:
    mqtt_client_path = abspath(join(dirname(__file__), '..', '..', 'tb_mqtt_client'))
    from os import environ
    if exists(mqtt_client_path) and TBUtility.str_to_bool(environ.get(DEV_MODE_PARAMETER_NAME, 'false')):
        path.insert(0, mqtt_client_path)
        from tb_gateway_mqtt import TBGatewayMqttClient, TBDeviceMqttClient, \
            GATEWAY_ATTRIBUTES_RESPONSE_TOPIC, GATEWAY_ATTRIBUTES_TOPIC
        import tb_device_mqtt
    else:
        print("tb-mqtt-client library not found - installing...")
        TBUtility.install_package('tb-mqtt-client')
        from tb_gateway_mqtt import TBGatewayMqttClient, TBDeviceMqttClient, \
            GATEWAY_ATTRIBUTES_RESPONSE_TOPIC, GATEWAY_ATTRIBUTES_TOPIC
        import tb_device_mqtt

tb_device_mqtt.DEFAULT_TIMEOUT = 3


class TBClient(threading.Thread):
    def __init__(self, config, config_folder_path, logger):
        self.__logger = logger
        super().__init__()
        self.name = 'Connection thread.'
        self.daemon = True
        self.__config_folder_path = config_folder_path
        self.__config = config
        self.__host = config["host"]
        self.__port = config.get("port", 1883)
        self.__proxy_host = config.get("proxy_host", None)
        self.__proxy_port = config.get("proxy_port", None)
        self.__default_quality_of_service = config.get("qos", 1)
        self.__min_reconnect_delay = 1
        self.client: Union[TBGatewayMqttClient, None] = None
        self.__ca_cert = None
        self.__private_key = None
        self.__cert = None
        self.__client_id = ""
        self.__username = None
        self.__password = None
        self.__is_connected = False
        self.__stopped = False
        self.__stop_event = threading.Event()
        self.__paused = False
        # TODO: remove this flag
        self.__initial_connection_done = False
        self._last_cert_check_time = 0
        self.__service_subscription_callbacks = []

        # check if provided creds or provisioning strategy
        provisioning_configuration = TBUtility.get_provisioning_configuration_from_envs()
        if config.get('security') and not provisioning_configuration:
            self._create_mqtt_client(config['security'])
        elif config.get('provisioning') or provisioning_configuration:
            creds = None
            try:
                if exists(self.__config_folder_path + PROVISIONED_CREDENTIALS_FILENAME):
                        with open(self.__config_folder_path + PROVISIONED_CREDENTIALS_FILENAME, 'r') as file:
                            credentials = load(file)
                        creds = self._get_provisioned_creds(credentials)
            except Exception as e:
                self.__logger.error('Error reading provisioned credentials: %s', e)
            if not creds:
                credentials = config.get('provisioning', provisioning_configuration)
                logger.info('Starting provisioning gateway...')

                credentials_type = credentials.pop('type', 'ACCESS_TOKEN')
                if credentials_type.upper() == 'ACCESS_TOKEN':
                    credentials['access_token'] = ''.join(random.choice(string.ascii_lowercase) for _ in range(15))
                elif credentials_type.upper() == 'MQTT_BASIC':
                    credentials['client_id'] = ''.join(random.choice(string.ascii_lowercase) for _ in range(15))
                    credentials['username'] = ''.join(random.choice(string.ascii_lowercase) for _ in range(15))
                    credentials['password'] = ''.join(random.choice(string.ascii_lowercase) for _ in range(15))
                elif credentials_type.upper() == 'X509_CERTIFICATE':
                    self._ca_cert_name = credentials.pop('caCert')
                    new_cert_path = self.__config_folder_path + 'cert.pem'
                    new_private_key_path = self.__config_folder_path + 'key.pem'
                    gen_hash = TBUtility.generate_certificate(new_cert_path, new_private_key_path).decode('utf-8')
                    credentials['hash'] = gen_hash
                else:
                    raise RuntimeError('Unknown provisioning type '
                                       '(Available options: ACCESS_TOKEN, MQTT_BASIC, X509_CERTIFICATE)')

                gateway_name = credentials.pop('deviceName',
                                               'Gateway ' + ''.join(random.choice(string.ascii_lowercase) for _ in range(5)))
                prov_gateway_key = credentials.pop('provisionDeviceKey')
                prov_gateway_secret = credentials.pop('provisionDeviceSecret')
                creds = TBDeviceMqttClient.provision(host=self.__host,
                                                     port=self.__port,
                                                     device_name=gateway_name,
                                                     provision_device_key=prov_gateway_key,
                                                     provision_device_secret=prov_gateway_secret,
                                                     gateway=True,
                                                     **credentials)

                if not creds:
                    logger.error('Provisioning failed, check your credentials and try again')
                    raise RuntimeError('Provisioning failed, check your credentials and try again')
                with open(self.__config_folder_path + PROVISIONED_CREDENTIALS_FILENAME, 'w') as file:
                    if hasattr(self, '_ca_cert_name') and self._ca_cert_name:
                        creds['caCert'] = self._ca_cert_name
                    file.writelines(dumps(creds))
                logger.info('Gateway provisioned')

                creds = self._get_provisioned_creds(creds)

            self._create_mqtt_client(creds)
        else:
            raise RuntimeError('Security section not provided')

        # pylint: disable=protected-access
        # Adding callbacks
        self.client._client._on_connect = self._on_connect  # noqa pylint: disable=protected-access
        self.client._client._on_disconnect = self._on_disconnect  # noqa pylint: disable=protected-access
        # self.client._client._on_log = self._on_log  # noqa pylint: disable=protected-access
        self.start()

    # def _on_log(self, *args):
    #     if "exception" in args[-1]:
    #         self.__logger.exception(args)
    #     else:
    #         self.__logger.debug(args)

    def _create_mqtt_client(self, credentials):
        env_vars = TBUtility.get_service_environmental_variables()
        self.__tls = bool((credentials.get('tls', False) or
                          credentials.get('caCert', False)))
        if env_vars.get('caCert') is not None:
            self.__tls = True
        if env_vars.get('type') is not None:
            credentials_type = env_vars.get('type')
        else:
            credentials_type = credentials.get('type')
        if credentials_type is None and credentials.get('accessToken') is not None:
            credentials_type = 'accessToken'
        if credentials_type == 'accessToken':
            if env_vars.get('accessToken') is not None:
                self.__username = str(env_vars["accessToken"])
            else:
                self.__username = str(credentials["accessToken"])
        if credentials_type == 'usernamePassword':
            if credentials.get("username") is not None:
                if env_vars.get('username') is not None:
                    self.__username = str(env_vars["username"])
                else:
                    self.__username = str(credentials["username"])
            if credentials.get("password") is not None:
                if env_vars.get('password') is not None:
                    self.__password = str(env_vars["password"])
                else:
                    self.__password = str(credentials["password"])
            if credentials.get("clientId") is not None:
                if env_vars.get('clientId') is not None:
                    self.__client_id = str(env_vars["clientId"])
                else:
                    self.__client_id = str(credentials["clientId"])

        rate_limits_config = self.__get_rate_limit_config()

        if rate_limits_config:
            self.client = TBGatewayMqttClient(self.__host, self.__port, self.__username, self.__password, self,
                                              quality_of_service=self.__default_quality_of_service,
                                              client_id=self.__client_id, **rate_limits_config)
        else:
            self.client = TBGatewayMqttClient(self.__host, self.__port, self.__username, self.__password, self,
                                              quality_of_service=self.__default_quality_of_service,
                                              client_id=self.__client_id)

        self.__ca_cert = self.__get_path_to_cert(credentials.get("caCert")) \
            if credentials.get("caCert") is not None else None
        if env_vars.get('caCert') is not None:
            self.__ca_cert = self.__get_path_to_cert(env_vars.get("caCert"))
        self.__private_key = self.__get_path_to_cert(credentials.get("privateKey")) \
            if credentials.get("privateKey") is not None else None
        if env_vars.get('privateKey') is not None:
            self.__private_key = self.__get_path_to_cert(env_vars.get("privateKey"))
        self.__cert = self.__get_path_to_cert(credentials.get("cert")) \
            if credentials.get("cert") is not None else None
        if env_vars.get('cert') is not None:
            self.__cert = self.__get_path_to_cert(env_vars.get("cert"))
        self.__check_cert_period = credentials.get('checkCertPeriod', 86400)
        self.__certificate_days_left = credentials.get('certificateDaysLeft', 3)

        if self.__tls:
            # check certificates for end date
            self._check_cert_thread = threading.Thread(name='Check Certificates Thread',
                                                       target=self._check_certificates,
                                                       daemon=True)
            self._check_cert_thread.start()

        cert_required = CERT_REQUIRED if (self.__ca_cert and
                                          self.__cert) else ssl.CERT_OPTIONAL if self.__cert else ssl.CERT_NONE

        self.client._client.tls_set(ca_certs=self.__ca_cert,
                                    certfile=self.__cert,
                                    keyfile=self.__private_key,
                                    tls_version=ssl.PROTOCOL_TLSv1_2,
                                    cert_reqs=cert_required,
                                    ciphers=None)  # noqa pylint: disable=protected-access
        if credentials.get("insecure", False):
            self.client._client.tls_insecure_set(True)  # noqa pylint: disable=protected-access
        if self.__logger.isEnabledFor(10):
            self.client._client.enable_logger(self.__logger)  # noqa pylint: disable=protected-access
        if self.__proxy_host is not None:
            if self.__proxy_port is not None:
                self.client._client.proxy_set(proxy_type=socks.HTTP, proxy_addr=self.__proxy_host, proxy_port=self.__proxy_port)
            else:
                self.client._client.proxy_set(proxy_type=socks.HTTP, proxy_addr=self.__proxy_host)

    def __get_rate_limit_config(self):
        rate_limits_config = {}
        if self.__config.get('messagesRateLimits'):
            rate_limits_config['messages_rate_limit'] = self.__config['messagesRateLimits']
        if self.__config.get('telemetryRateLimits'):
            rate_limits_config['telemetry_rate_limit'] = self.__config['rateLimits']
        if self.__config.get('telemetryDpRateLimits'):
            rate_limits_config['telemetry_dp_rate_limit'] = self.__config['dpRateLimits']

        if self.__config.get('deviceMessagesRateLimits'):
            rate_limits_config['device_messages_rate_limit'] = self.__config['deviceMessagesRateLimits']
        if self.__config.get('deviceTelemetryRateLimits'):
            rate_limits_config['device_telemetry_rate_limit'] = self.__config['deviceRateLimits']
        if self.__config.get('deviceTelemetryDpRateLimits'):
            rate_limits_config['device_telemetry_dp_rate_limit'] = self.__config['deviceDpRateLimits']

        if 'rate_limit' in inspect.signature(TBGatewayMqttClient.__init__).parameters:
            rate_limits_config = {}
            if self.__config.get('rateLimits'):
                rate_limits_config['rate_limit'] = 'DEFAULT_RATE_LIMIT' if self.__config.get(
                    'rateLimits') == 'DEFAULT_TELEMETRY_RATE_LIMIT' else self.__config['rateLimits']
            if ('dp_rate_limit' in inspect.signature(TBGatewayMqttClient.__init__).parameters and
                    self.__config.get('dpRateLimits')):
                rate_limits_config['dp_rate_limit'] = 'DEFAULT_RATE_LIMIT' if self.__config[
                    'dpRateLimits'] == 'DEFAULT_TELEMETRY_DP_RATE_LIMIT' else self.__config['dpRateLimits']

        return rate_limits_config

    def __get_path_to_cert(self, filename):
        if exists(self.__config_folder_path + filename):
            return self.__config_folder_path + filename
        else:
            return filename

    @staticmethod
    def _get_provisioned_creds(credentials):
        creds = {}
        if credentials.get('credentialsType') == 'ACCESS_TOKEN':
            creds['accessToken'] = credentials['credentialsValue']
            creds['type'] = 'accessToken'
        elif credentials.get('credentialsType') == 'MQTT_BASIC':
            creds['clientId'] = credentials['credentialsValue']['clientId']
            creds['username'] = credentials['credentialsValue']['userName']
            creds['password'] = credentials['credentialsValue']['password']
            creds['type'] = 'usernamePassword'
        elif credentials.get('credentialsType') == 'X509_CERTIFICATE':
            creds['tls'] = True
            creds['caCert'] = credentials['caCert']
            creds['privateKey'] = 'key.pem'
            creds['cert'] = 'cert.pem'
            creds['type'] = 'x509Certificate'

        return creds

    def _check_certificates(self):
        while not self.__stopped and not self.__paused:
            if time() - self._last_cert_check_time >= self.__check_cert_period:
                if self.__cert:
                    self.__logger.info('Will generate new certificate')
                    new_cert = TBUtility.check_certificate(self.__cert, key=self.__private_key,
                                                           days_left=self.__certificate_days_left)

                    if new_cert:
                        self.client.send_attributes({'newCertificate': new_cert})

                if self.__ca_cert:
                    is_outdated = TBUtility.check_certificate(self.__ca_cert, generate_new=False,
                                                              days_left=self.__certificate_days_left)

                    if is_outdated:
                        self.client.send_attributes({'CACertificate': 'CA certificate will outdated soon'})

                self._last_cert_check_time = time()

            sleep(10)

    def pause(self):
        self.__paused = True

    def unpause(self):
        self.__paused = False

    def is_connected(self):
        return self.__is_connected and self.client.rate_limits_received

    def _on_connect(self, client, userdata, flags, result_code, parameters, *extra_params):
        self.__logger.info('MQTT client connected to platform %s: %s', self.__host, self.__port)
        self.__logger.debug('MQTT client %r connected to platform', str(client))
        if ((isinstance(result_code, int) and result_code == 0) or
                (hasattr(result_code, 'value') and result_code.value == 0)):
            self.__is_connected = True
            if self.__initial_connection_done:
                self.client.rate_limits_received = True  # Added to avoid stuck on reconnect, if rate limits reached,
                # TODO: move to high priority.
            else:
                self.__initial_connection_done = True
        # pylint: disable=protected-access
        self.client._on_connect(client, userdata, flags, result_code, parameters, *extra_params) # noqa pylint: disable=protected-access
        try:
            if isinstance(result_code, int) and result_code != 0:
                if result_code in (159, 151):
                    self.__logger.warning("Connection rate exceeded.")
            else:
                if result_code.getName().lower() == "connection rate exceeded":
                    self.__logger.warning("Connection rate exceeded.")
        except Exception as e:
            self.__logger.exception("Error in on_connect callback: %s", exc_info=e)

        for callback in self.__service_subscription_callbacks:
            callback()

    def _on_disconnect(self, client, userdata, disconnect_flags, result_code, properties=None):
        # pylint: disable=protected-access
        self.__is_connected = False
        if self.client._client != client: # noqa pylint: disable=protected-access
            self.__logger.info("TB client %s has been disconnected. Current client for connection is: %s", str(client),
                               str(self.client._client)) # noqa pylint: disable=protected-access
            client.loop_stop()
        else:
            self.__is_connected = False
            if len(inspect.signature(self.client._on_disconnect).parameters) == 5: # noqa pylint: disable=protected-access
                self.client._on_disconnect(client, userdata, disconnect_flags, result_code, properties) # noqa pylint: disable=protected-access
            elif len(inspect.signature(self.client._on_disconnect).parameters) == 4: # noqa pylint: disable=protected-access
                self.client._on_disconnect(client, userdata, result_code, properties) # noqa pylint: disable=protected-access
            else:
                self.client._on_disconnect(client, userdata, result_code) # noqa pylint: disable=protected-access

    def stop(self):
        # self.disconnect()
        self.client.stop()
        self.__stopped = True
        self.__stop_event.set()

    def disconnect(self):
        self.__paused = True
        self.unsubscribe('*')
        return self.client.disconnect()

    def unsubscribe(self, subscription_id):
        self.client.gw_unsubscribe(subscription_id)
        self.client.unsubscribe_from_attribute(subscription_id)

    def connect(self, min_reconnect_delay=5, timeout=None):
        self.__paused = False
        self.__stopped = False
        self.__min_reconnect_delay = max(min_reconnect_delay, 2)

        start_connect_time = time()
        previous_connection_time = time()
        first_connection = True

        try:
            while not self.client.is_connected() and not self.__stopped:
                if not self.__paused:
                    if self.__stopped:
                        break
                    if timeout is not None and time() - start_connect_time > timeout:
                        self.__logger.error("Connection timeout.")
                        break
                    if self.client._client.is_connected():
                        self.__logger.info("Client connected to platform.")
                        if self.client.rate_limits_received:
                            self.__logger.info("Rate limits received.")
                        else:
                            self.__logger.info("Waiting for rate limits configuration...")
                    else:
                        self.__logger.info("Connecting to ThingsBoard...")
                    try:
                        if first_connection or time() - previous_connection_time > min_reconnect_delay:
                            first_connection = False
                            self.__send_connect()
                            previous_connection_time = time()
                        else:
                            sleep(1)
                    except ConnectionRefusedError:
                        self.__logger.error("Connection refused.")
                    except ConnectionError as e:
                        self.__logger.error("Connection error, cannot connect with error: %s", e)
                    except (ssl.SSLEOFError, ssl.SSLZeroReturnError) as e:
                        self.__logger.warning("Cannot use TLS connection on this port. "
                                              "Client will try to connect without TLS.")
                        self.__logger.debug("Error: %s", e)
                        self.client._client._ssl = False # noqa pylint: disable=protected-access
                        try:
                            self.__send_connect()
                        except Exception:
                            continue
                    except Exception as e:
                        self.__logger.error("Error in connect: %s", exc_info=e)
                sleep(1)
        except Exception as e:
            self.__logger.exception(e)
            sleep(10)

    def __send_connect(self):
        self.__logger.info(f"Sending connect to {self.__host}:{self.__port}")
        try:
            self.client.connect(keepalive=self.__config.get("keep_alive", 120),
                                min_reconnect_delay=self.__min_reconnect_delay)
            self.__logger.debug("Connect msg sent to platform.")
        except socket.gaierror as e:
            raise ConnectionError(e) from e

    def run(self):
        while not self.__stopped:
            try:
                if self.__stop_event.wait(timeout=0.1):
                    break
            except KeyboardInterrupt:
                self.__stopped = True
                self.__stop_event.set()
            except TimeoutError:
                sleep(0.1)
            except Exception as e:
                self.__logger.exception("Error in main MQTT client loop: %s", exc_info=e)

    def get_config_folder_path(self):
        return self.__config_folder_path

    def register_service_subscription_callback(self, subscribe_to_required_topics):
        self.__service_subscription_callbacks.append(subscribe_to_required_topics)

    @property
    def config(self):
        return self.__config

    def is_stopped(self):
        return self.__stopped

    def get_max_payload_size(self):
        return self.client.max_payload_size # noqa pylint: disable=protected-access

    def update_logger(self):
        self.__logger.setLevel(getLogger("tb_connection").level)
        self.__logger.handlers = getLogger("tb_connection").handlers
        self.__logger.manager = getLogger("tb_connection").manager
        self.__logger.disabled = getLogger("tb_connection").disabled
        self.__logger.filters = getLogger("tb_connection").filters
        self.__logger.propagate = getLogger("tb_connection").propagate
        self.__logger.parent = getLogger("tb_connection").parent
        tb_device_mqtt.log = self.__logger

    def get_rate_limits(self):
        return {
            'devices_connected_through_gateway_telemetry_messages_rate_limit': deepcopy(self.client._devices_connected_through_gateway_telemetry_messages_rate_limit.__dict__),  # noqa
            'devices_connected_through_gateway_telemetry_datapoints_rate_limit': deepcopy(self.client._devices_connected_through_gateway_telemetry_datapoints_rate_limit.__dict__),  # noqa
            'devices_connected_through_gateway_messages_rate_limit': deepcopy(self.client._devices_connected_through_gateway_messages_rate_limit.__dict__),  # noqa
            'messages_rate_limit': deepcopy(self.client._messages_rate_limit.__dict__),  # noqa
            'telemetry_rate_limit': deepcopy(self.client._telemetry_rate_limit.__dict__),  # noqa
            'telemetry_dp_rate_limit': deepcopy(self.client._telemetry_dp_rate_limit.__dict__),  # noqa
            'max_inflight_messages': self.client._client._max_inflight_messages,  # noqa
            'max_payload_size': self.client.max_payload_size
        }

    def is_subscribed_to_service_attributes(self):
        return GATEWAY_ATTRIBUTES_RESPONSE_TOPIC in self.client._gw_subscriptions.values() and GATEWAY_ATTRIBUTES_TOPIC in self.client._gw_subscriptions.values() # noqa pylint: disable=protected-access

    def update_client_rate_limits_with_previous(self, previous_limits):
        self.client._devices_connected_through_gateway_telemetry_messages_rate_limit = tb_device_mqtt.RateLimit(
            previous_limits.get('devices_connected_through_gateway_telemetry_messages_rate_limit', {}))
        self.client._devices_connected_through_gateway_telemetry_datapoints_rate_limit = tb_device_mqtt.RateLimit(
            previous_limits.get('devices_connected_through_gateway_telemetry_datapoints_rate_limit', {}))
        self.client._devices_connected_through_gateway_messages_rate_limit = tb_device_mqtt.RateLimit(
            previous_limits.get('devices_connected_through_gateway_messages_rate_limit', {}))

        self.client._messages_rate_limit = tb_device_mqtt.RateLimit(previous_limits.get('messages_rate_limit', {})) # noqa pylint: disable=protected-access
        self.client._telemetry_rate_limit = tb_device_mqtt.RateLimit(previous_limits.get('telemetry_rate_limit', {})) # noqa pylint: disable=protected-access
        self.client._telemetry_dp_rate_limit = tb_device_mqtt.RateLimit(previous_limits.get('telemetry_dp_rate_limit', {})) # noqa pylint: disable=protected-access
        self.client.max_inflight_messages_set(previous_limits.get('max_inflight_messages', 20)) # noqa pylint: disable=protected-access
        self.client.max_payload_size = previous_limits.get('max_payload_size', 8196) # noqa pylint: disable=protected-access



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from copy import copy
from gc import collect
from logging import getLogger
from os import path, makedirs, remove
from queue import Queue, Full
from sqlite3 import ProgrammingError, DatabaseError
from threading import Event, Thread
from time import sleep, monotonic

from thingsboard_gateway.storage.event_storage import EventStorage
from thingsboard_gateway.storage.sqlite.database import Database
from thingsboard_gateway.storage.sqlite.database_connector import DatabaseConnector
from thingsboard_gateway.storage.sqlite.sqlite_event_storage_pointer import Pointer
from thingsboard_gateway.storage.sqlite.storage_settings import StorageSettings


class SQLiteEventStorage(EventStorage):

    def __init__(self, config, logger, main_stop_event):
        super().__init__(config, logger, main_stop_event)
        self.__log = logger
        self.__log.info("Sqlite Storage initializing...")
        self.write_queue = Queue(-1)
        self.stopped = Event()

        self.__settings = config if isinstance(config, StorageSettings) else StorageSettings(config)
        if self.__settings.warnings:
            for warning in self.__settings.warnings:
                self.__log.warning("%s", str(warning))

        self.__ensure_data_folder_exists()
        self.__pointer = Pointer(self.__settings.data_file_path, log=self.__log)
        self.__default_database_name = self.__settings.db_file_name
        self.__is_max_db_amount_reached = False
        self.__read_database_name, self.__write_database_name = (
            self.__select_initial_db_files()
        )

        read_database_path = str(
            path.join(self.__settings.directory_path, self.__read_database_name)
        )
        write_database_path = str(
            path.join(self.__settings.directory_path, self.__write_database_name)
        )
        override_write_settings = copy(self.__settings)
        self.update_settings(
            storage_settings=override_write_settings,
            data_file_path=write_database_path,
        )

        self.__write_database = Database(
            override_write_settings,
            self.write_queue,
            self.__log,
            stopped=self.stopped,
            should_read=False,
            should_write=True,
            on_rotate_callback=self.on_write_database_callback

        )
        self.__write_database.start()
        self.__log.debug("Write DB thread started for %s", self.__write_database_name)

        self.__write_database.init_table()

        override_read_db_settings = copy(self.__settings)
        self.update_settings(
            storage_settings=override_read_db_settings,
            data_file_path=read_database_path,
        )

        if self.__read_database_name == self.__write_database_name:
            self.__read_database = self.__write_database
            self.__read_database.should_read = True
        else:

            self.__read_database = Database(
                override_read_db_settings,
                self.write_queue,
                self.__log,
                stopped=self.stopped,
                should_read=True,
                should_write=False,
            )

            self.__read_database.start()
            self.__log.debug("Read DB thread started for %s", self.__read_database_name)

        self.__log.debug(
            "Sqlite storage initialized (read=%s, write=%s)",
            self.__read_database_name,
            self.__write_database_name,
        )
        self._database_files = self.__pointer.sort_db_files()

        if not self.__read_database.database_has_records():
            self.__read_database.process_file_limit()
            if self.__read_database.reached_size_limit:
                self.__log.debug("Initial read DB oversize & empty → rotating")
                self.__rotate_read_database()
        if not self.__read_database.database_has_records() and len(self._database_files) > 1:
            self.__rotate_read_database()
        self.delete_time_point = 0
        self.__join_thread_timeout = 5
        self.__event_pack_processing_start = monotonic()

    def __select_initial_db_files(self):
        all_db_files = self.__pointer.sort_db_files()
        if len(all_db_files) == 1 and self.__default_database_name < all_db_files[0]:
            return all_db_files[0], all_db_files[0]
        if len(all_db_files) > 1:
            return all_db_files[0], all_db_files[-1]
        return self.__default_database_name, self.__default_database_name

    def __ensure_data_folder_exists(self):
        data_path = self.__settings.data_file_path
        if not path.exists(data_path):
            directory = path.dirname(data_path)
            if not path.exists(directory):
                self.__log.debug("Data folder missing, creating %s", directory)
                try:
                    makedirs(directory)
                    self.__log.info("Created directory %s", directory)
                except OSError as e:
                    self.__log.error("OS error creating %s: %s", directory, e)
                    self.__log.debug("OS stack:", exc_info=e)
                except Exception as e:
                    self.__log.exception("Unexpected error creating %s", directory)
                    self.__log.debug("Stack:", exc_info=e)

    def __create_read_database(self):
        read_database_filename = self._database_files[0]
        full_path = str(
            path.join(self.__settings.directory_path, read_database_filename)
        )

        rotate_read_database_settings = copy(self.__settings)
        self.update_settings(
            storage_settings=rotate_read_database_settings, data_file_path=full_path
        )

        try:
            self.__read_database = Database(
                rotate_read_database_settings,
                self.write_queue,
                self.__log,
                stopped=self.stopped,
                should_read=True,
                should_write=False,
            )
            self.__read_database.start()
            self.__log.debug("Switched read DB to %s", read_database_filename)

        except Exception:
            self.__log.exception("Failed to start read DB %s", read_database_filename)

    def __cleanup_write_db_after_thread_termination(self) -> None:
        if self.__read_database != self.__write_database:
            timeout = 2.0
            start = monotonic()
            while (
                    not self.__write_database.process_queue.empty()
                    and monotonic() - start < timeout
            ):
                sleep(0.1)
            try:
                self.__finalize_write_database_thread()
                self.__log.debug("Oversize DB cleaned up")
            except RuntimeError as e:
                self.__log.debug("Thread error during oversize cleanup: %s", e)
            except DatabaseError as e:
                self.__log.debug("DB error during oversize cleanup: %s", e)
            except Exception:
                self.__log.debug("Unexpected error cleaning oversize DB")
            finally:
                self.__write_database = None
        else:
            self.__log.trace("Oversize cleanup skipped (read==write)")
            self.__read_database.should_write = False

    def __finalize_write_database_thread(self) -> None:
        self.__write_database.db.commit()
        try:
            self.__write_database.interrupt()
            self.__log.trace("Interrupted oversize DB thread")
        except AttributeError:
            pass
        try:
            self.__write_database.join(timeout=self.__join_thread_timeout)
            if self.__write_database.is_alive():
                self.__log.warning("DB thread alive after join timeout")
        except RuntimeError as e:
            self.__log.debug("Join runtime error: %s", e)
        except Exception as e:
            self.__log.debug("Join error: %s", e)
        try:
            self.__write_database.close_db()
            self.__log.trace("Closed oversize DB connection")
        except ProgrammingError as e:
            self.__log.debug("Close_db on closed DB: %s", e)
        except Exception as e:
            self.__log.debug("Close_db error: %s", e)
        self.__write_database.db.close()

    def __check_and_handle_max_db_count(self) -> bool:
        count = len(self._database_files)
        if count >= self.__settings.max_db_amount:
            self.__is_max_db_amount_reached = True
            self.__log.warning("Max DB count (%d) reached—dropping writes", count)
            return True
        if self.__is_max_db_amount_reached:
            self.__log.debug("DB count cleared, creating new write DB")
            self.__is_max_db_amount_reached = False
            new_config_after_max_db_count = self.__prepare_new_db_configuration()
            self.__start_write_database(new_config=new_config_after_max_db_count)
        return False

    def __rotate_read_database(self):
        self.__log.debug("Rotate after read completion")
        if path.exists(self.__read_database.settings.data_file_path):
            self.__delete_db_files(self.__read_database.settings.data_file_path)
        self.delete_time_point = 0

        if len(self._database_files) > 1:
            self.__create_read_database()
        else:
            self.__log.trace("Only one DB left, reusing write DB")
            self.__read_database = self.__write_database
            self.__read_database.should_read = True

        self.__check_and_handle_max_db_count()

    def event_pack_processing_done(self):
        self.__log.trace(
            "Batch done in %d ms",
            int((monotonic() - self.__event_pack_processing_start) * 1000),
        )
        if not self.stopped.is_set():
            self.delete_data(self.delete_time_point)
            if not self.__read_database.database_has_records():

                self.__read_database.process_file_limit()
                if self.__read_database.reached_size_limit:
                    self.__rotate_read_database()

        collect()

    def get_event_pack(self):
        if not self.stopped.is_set():
            self.__event_pack_processing_start = monotonic()
            event_pack_messages = []
            data_from_storage = self.read_data()
            if not data_from_storage and not path.exists(
                    self.__read_database.settings.data_file_path
            ):
                self.__rotate_read_database()
            if not data_from_storage and len(
                    self._database_files) > 1 and not self.__read_database.database_has_records():
                self.__rotate_read_database()
            event_pack_messages = self.process_event_storage_data(
                data_from_storage=data_from_storage,
                event_pack_messages=event_pack_messages,
            )

            if event_pack_messages:
                self.__log.trace(
                    "Retrieved %r records from storage in %r ms, left in storage: %r",
                    len(event_pack_messages),
                    int((monotonic() - self.__event_pack_processing_start) * 1000),
                    self.__read_database.get_stored_messages_count(),
                )

            self.__read_database.can_prepare_new_batch()

            return event_pack_messages

        else:
            return []

    def process_event_storage_data(self, data_from_storage, event_pack_messages):

        if not data_from_storage:
            return []
        for row in data_from_storage:
            try:
                if not row:
                    return []
                element_to_insert = row["message"]
                if not element_to_insert:
                    continue

                event_pack_messages.append(element_to_insert)
                if not self.delete_time_point or self.delete_time_point < row["id"]:
                    self.delete_time_point = row["id"]
            except (IndexError, KeyError) as e:

                self.__log.error(
                    "Failed to extract message from storage row %r: %s", row, e, )
                self.__log.debug("Stack:", exc_info=e)
                continue
            except Exception as e:
                self.__log.error(
                    "Error occurred while reading data from storage: %s", e
                )

        return event_pack_messages

    def __finalize_read_database_thread(self):
        try:
            timeout = 2.0
            start = monotonic()
            while (
                    not self.__read_database.process_queue.empty()
                    and monotonic() - start < timeout
            ):
                sleep(0.05)
            self.__read_database.db.commit()
            self.__read_database.interrupt()
            self.__read_database.join(timeout=self.__join_thread_timeout)
            self.__read_database.close_db()
            self.__read_database.db.close()
        except Exception:
            self.__log.debug("Interruption during delete cleanup")
            sleep(0.1)

    def __delete_db_files(self, path_to_db_file):

        self.__finalize_read_database_thread()
        deleted = False
        for suffix in ("", "-shm", "-wal"):
            full = path_to_db_file + suffix
            if path.exists(full):
                try:
                    remove(full)
                    deleted = True
                except Exception as e:
                    self.__log.exception("Failed delete %s: %s", full, e)
        if deleted:
            self._database_files.remove(self.__read_database.settings.db_file_name)
        if not deleted:
            self.__log.error("No DB files to delete under %s", path_to_db_file)

    def read_data(self):
        return self.__read_database.read_data()

    def delete_data(self, row_id):
        return self.__read_database.delete_data(row_id=row_id)

    def __prepare_new_db_configuration(self) -> StorageSettings:
        new_db_name = self.__pointer.generate_new_file_name()
        data_file_path = str(path.join(self.__settings.directory_path, new_db_name))

        created_database_settings = copy(self.__settings)
        self.update_settings(
            storage_settings=created_database_settings, data_file_path=data_file_path
        )

        return created_database_settings

    def __start_write_database(self, new_config: StorageSettings):
        self.__write_database = Database(
            new_config,
            self.write_queue,
            self.__log,
            stopped=self.stopped,
            should_read=False,
            should_write=True,
            on_rotate_callback=self.on_write_database_callback,
        )
        self.__write_database.start()
        self.__log.debug(
            "Started new write DB %s", self.__write_database.settings.db_file_name
        )
        self._database_files.append(self.__write_database.settings.db_file_name)

    def on_write_database_callback(self):
        if self.__write_database.reached_size_limit and not len(self._database_files) >= self.__settings.max_db_amount:
            self.__log.debug("Write DB %s reached limit",
                             self.__write_database.settings.db_file_name, )
            new_write_database_config = self.__prepare_new_db_configuration()
            self.__cleanup_write_db_after_thread_termination()
            self.__start_write_database(new_config=new_write_database_config)

    def put(self, message):
        try:

            if self.__is_max_db_amount_reached:
                return False
            if not self.stopped.is_set():
                if self.__write_database is None or self.__write_database.reached_size_limit and self.__check_and_handle_max_db_count():
                    return False
                self.__log.trace("Queuing message: %r", message)
                self.write_queue.put_nowait(message)
                return True
            return False
        except Full:
            self.__log.error("Storage queue full—dropped message")
            return False
        except Exception as e:
            self.__log.exception("Failed to put message, %s", e)

    def stop(self):
        self.stopped.set()
        self.__read_database.close_db()
        self.__write_database.close_db()
        collect()

    def len(self):
        write_queue_size = self.write_queue.qsize()
        read_database_stored_messages_count = (
            self.__read_database.get_stored_messages_count()
        )
        saved_databases_rows_count = 0
        write_database_stored_messages_count = 0

        if (
                self.__write_database is not None
                and self.__write_database != self.__read_database
        ):
            write_database_stored_messages_count = (
                self.__write_database.get_stored_messages_count()
            )

        if len(self._database_files) > 2:
            result = {"result": 0}
            check_thread = Thread(
                target=self.__check_db_sizes,
                args=(result,),
                daemon=True,
                name="Database row count check thread",
            )
            check_thread.start()
            try:
                check_thread.join(timeout=self.__join_thread_timeout)
                saved_databases_rows_count = result["result"]
            except Exception as e:
                self.__log.error("Database row count check thread failed: %s", e)
                self.__log.debug("Database row count check thread failed:", exc_info=e)

        return (
                write_queue_size
                + read_database_stored_messages_count
                + write_database_stored_messages_count
                + saved_databases_rows_count
        )

    def __check_db_sizes(self, result):
        databases_rows_count = 0
        for database_name in self._database_files:
            if database_name not in (
                    self.__read_database.settings.db_file_name,
                    self.__write_database.settings.db_file_name,
            ):
                try:
                    db_path = path.join(self.__settings.directory_path, database_name)
                    db_connector = DatabaseConnector(
                        db_path,
                        self.__log,
                        self._main_stop_event,
                    )
                    db_connector.connect_on_closed_db(database_path=db_path)
                    cursor = db_connector.closed_db_connection.cursor()
                    cursor.execute("SELECT COUNT(id) FROM messages;")

                    row = cursor.fetchone()
                    if row:
                        databases_rows_count += row[0]

                except Exception as e:
                    self.__log.error(
                        "Failed to check db size for %s: %s", database_name, e
                    )
                    self.__log.debug("Stack trace:", exc_info=e)
                    continue

        result["result"] = databases_rows_count

    @staticmethod
    def update_settings(storage_settings: StorageSettings, data_file_path: str):
        storage_settings.data_file_path = data_file_path
        storage_settings.directory_path = path.dirname(data_file_path)
        storage_settings.db_file_name = path.basename(data_file_path)

    def update_logger(self):
        self.__log = getLogger("storage")



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

DEVICE_OBJECT_TYPE_CODE = '8'
OBJECT_TYPE_PARAMETER = 'object-type'
OBJECT_INSTANCE_PARAMETER = 'object-instance'
DEVICE_OBJ_INSTANCE_PARAMETER = 'device obj.-instance'
OBJECT_NAME_PARAMETER = 'object-name'
DEFAULT_POLL_PERIOD = 10000
EDE_FILE_PATH_PARAMETER = 'edeFilePath'
EDE_CONFIG_PARAMETER = 'edeConfig'
ALLOWED_APDU = (50, 128, 206, 480, 1024, 1476)
SUPPORTED_OBJECTS_TYPES = {
    '0': 'analogInput',
    '1': 'analogOutput',
    '2': 'analogValue',
    '3': 'binaryInput',
    '4': 'binaryOutput',
    '5': 'binaryValue',
}



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from thingsboard_gateway.connectors.bacnet.constants import (
    DEFAULT_POLL_PERIOD,
    OBJECT_INSTANCE_PARAMETER,
    OBJECT_NAME_PARAMETER,
    OBJECT_TYPE_PARAMETER,
    SUPPORTED_OBJECTS_TYPES
)


class EDEConnectorConfig:
    def __init__(self):
        self._config = {
            "devices": []
        }

    @property
    def config(self):
        return self._config

    def add_device(self, device_id, device_name, device_objects):
        timeseries_config = self._get_device_objects(device_objects)

        self._config['devices'].append({
            "deviceInfo": {
                "deviceNameExpression": f"{device_name}",
                "deviceProfileExpression": "default",
                "deviceNameExpressionSource": "constant",
                "deviceProfileExpressionSource": "constant"
            },
            "deviceId": device_id,
            "host": "*",
            "port": "*",
            "mask": "*",
            "pollPeriod": DEFAULT_POLL_PERIOD,
            "timeseries": timeseries_config
        })

    @staticmethod
    def _get_device_objects(device_objects):
        device_objects_config = []

        for device_obj in device_objects:
            try:
                device_objects_config.append({
                    "key": device_obj.get(OBJECT_NAME_PARAMETER, 'Unknown Object'),
                    "objectType": SUPPORTED_OBJECTS_TYPES[device_obj[OBJECT_TYPE_PARAMETER]],
                    "objectId": device_obj[OBJECT_INSTANCE_PARAMETER],
                    "propertyId": "presentValue"
                })
            except KeyError:
                continue

        return device_objects_config



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from datetime import timezone
from time import time

from asyncua.ua.uatypes import VariantType

from thingsboard_gateway.connectors.opcua.opcua_converter import OpcUaConverter
from thingsboard_gateway.gateway.constants import TELEMETRY_PARAMETER, ATTRIBUTES_PARAMETER, REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility

DATA_TYPES = {
    'attributes': ATTRIBUTES_PARAMETER,
    'timeseries': TELEMETRY_PARAMETER
}

VARIANT_TYPE_HANDLERS = {
    VariantType.ExtensionObject: lambda data: str(data),
    VariantType.DateTime: lambda data: data.replace(
        tzinfo=timezone.utc).isoformat() if data.tzinfo is None else data.isoformat(),
    VariantType.StatusCode: lambda data: data.name,
    VariantType.QualifiedName: lambda data: data.to_string(),
    VariantType.NodeId: lambda data: data.to_string(),
    VariantType.ExpandedNodeId: lambda data: data.to_string(),
    VariantType.ByteString: lambda data: data.hex(),
    VariantType.XmlElement: lambda data: data.decode('utf-8'),
    VariantType.Guid: lambda data: str(data),
    VariantType.DiagnosticInfo: lambda data: data.to_string(),
    VariantType.Null: lambda data: None
}

ERROR_MSG_TEMPLATE = "Bad status code: {} for node: {} with description {}"


class OpcUaUplinkConverter(OpcUaConverter):
    def __init__(self, config, logger):
        self._log = logger
        self.__config = config

    def process_datapoint(self, config, val, basic_timestamp, device_report_strategy):
        try:
            error = None
            data = val.Value.Value
            if isinstance(data, list):
                data = [str(item) for item in data]
            else:
                handler = VARIANT_TYPE_HANDLERS.get(val.Value.VariantType, lambda d: d if not hasattr(d, 'to_string') else d.to_string())
                data = handler(data)

            if data is None and val.StatusCode.is_bad():
                data = str.format(ERROR_MSG_TEMPLATE,val.StatusCode.name, val.data_type, val.StatusCode.doc)
                error = True

            timestamp_location = config.get('timestampLocation', 'gateway').lower()
            timestamp = basic_timestamp  # Default timestamp
            if timestamp_location == 'sourcetimestamp' and val.SourceTimestamp is not None:
                timestamp = val.SourceTimestamp.timestamp() * 1000
            elif timestamp_location == 'servertimestamp' and val.ServerTimestamp is not None:
                timestamp = val.ServerTimestamp.timestamp() * 1000

            section = DATA_TYPES[config['section']]
            datapoint_key = TBUtility.convert_key_to_datapoint_key(config['key'], device_report_strategy, config, self._log)
            if section == TELEMETRY_PARAMETER:
                return TelemetryEntry({datapoint_key: data}, ts=timestamp), error
            elif section == ATTRIBUTES_PARAMETER:
                return {datapoint_key: data}, error
        except Exception as e:
            return None, str(e)

    def convert(self, configs, values) -> ConvertedData:
        StatisticsService.count_connector_message(self._log.name, 'convertersMsgProcessed')
        basic_timestamp = int(time() * 1000)

        try:
            if not isinstance(configs, list):
                configs = [configs]
            if not isinstance(values, list):
                values = [values]

            converted_data = ConvertedData(device_name=self.__config['device_name'], device_type=self.__config['device_type'])

            device_report_strategy = None
            try:
                device_report_strategy = ReportStrategyConfig(self.__config.get(REPORT_STRATEGY_PARAMETER))
            except ValueError as e:
                self._log.trace("Report strategy config is not specified for device %s: %s", self.__config['device_name'], e)

            telemetry_batch = []
            attributes_batch = []

            for config, val in zip(configs, values):
                result, error = self.process_datapoint(config, val, basic_timestamp, device_report_strategy)
                if result is not None:
                    if isinstance(result, TelemetryEntry):
                        telemetry_batch.append(result)
                    elif isinstance(result, dict):
                        attributes_batch.append(result)

            converted_data.add_to_telemetry(telemetry_batch)
            for attr in attributes_batch:
                converted_data.add_to_attributes(attr)

            StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced', count=converted_data.attributes_datapoints_count)
            StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced', count=converted_data.telemetry_datapoints_count)

            return converted_data
        except Exception as e:
            self._log.exception("Error occurred while converting data: ", exc_info=e)
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')

    @staticmethod
    def fill_telemetry(results):
        telemetry_batch = []
        for result, error in results:
            if isinstance(result, TelemetryEntry):
                telemetry_batch.append(result)
        return telemetry_batch

    @staticmethod
    def fill_attributes(results):
        attributes_batch = []
        for result, error in results:
            if isinstance(result, dict):
                attributes_batch.append(result)
        return attributes_batch



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from enum import Enum

from thingsboard_gateway.gateway.constants import REPORT_PERIOD_PARAMETER, ReportStrategy, \
    TYPE_PARAMETER, AGGREGATION_FUNCTION_PARAMETER, TTL_PARAMETER, DEFAULT_REPORT_STRATEGY_CONFIG


class AggregationFunction(Enum):
    NONE = "NONE"
    SUM = "SUM"
    COUNT = "COUNT"
    MIN = "MIN"
    MAX = "MAX"
    AVERAGE = "AVERAGE"

    @classmethod
    def from_string(cls, value: str):
        for agg_function in cls:
            if agg_function.value.upper() == value.upper():
                return agg_function
        raise ValueError("Invalid aggregation function value: %r" % value)


class ReportStrategyConfig:
    __slots__ = ["report_period", "ttl", "report_strategy", "aggregation_function", "__hash"]

    def __init__(self, config, default_report_strategy_config=None):
        if default_report_strategy_config is None:
            default_report_strategy_config = {}
        if isinstance(config, ReportStrategyConfig):
            self.report_period = config.report_period
            self.ttl = config.ttl
            self.report_strategy = config.report_strategy
            self.aggregation_function = config.aggregation_function
            self.__hash = config.__hash
            return

        if config is None or not isinstance(config, dict):
            raise ValueError("Report strategy config is not specified")
        if (config.get(REPORT_PERIOD_PARAMETER) is None and config.get(TYPE_PARAMETER) is None
                and not default_report_strategy_config):
            raise ValueError("Report strategy config is not specified")
        if default_report_strategy_config:
            self.report_period = max(config.get(REPORT_PERIOD_PARAMETER,
                                                default_report_strategy_config[REPORT_PERIOD_PARAMETER]) - 10, 1)
            report_strategy_type = config.get(TYPE_PARAMETER, default_report_strategy_config[TYPE_PARAMETER])
        else:
            self.report_period = max(config.get(REPORT_PERIOD_PARAMETER, 0), 1)
            report_strategy_type = config.get(TYPE_PARAMETER)
        self.report_strategy = ReportStrategy.from_string(report_strategy_type)
        if self.report_strategy not in (ReportStrategy.ON_REPORT_PERIOD, ReportStrategy.ON_CHANGE_OR_REPORT_PERIOD):
            self.report_period = None
        self.aggregation_function = config.get(AGGREGATION_FUNCTION_PARAMETER)
        self.ttl = config.get(TTL_PARAMETER,
                              default_report_strategy_config.get(TTL_PARAMETER,
                                                                 DEFAULT_REPORT_STRATEGY_CONFIG[TTL_PARAMETER]))
        self.__validate_config()
        self.__hash = hash((self.report_period, self.report_strategy, self.aggregation_function))

    def __validate_config(self):
        if (self.report_strategy in (ReportStrategy.ON_REPORT_PERIOD, ReportStrategy.ON_CHANGE_OR_REPORT_PERIOD)
                and (self.report_period is None or self.report_period <= 0)):
            raise ValueError("Invalid report period value: %r" % str(self.report_period))
        if (self.aggregation_function is not None
                and self.aggregation_function not in AggregationFunction.__members__.values()):
            raise ValueError("Invalid aggregation function value: %r" % self.aggregation_function)

    def __hash__(self):
        return self.__hash

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (isinstance(other, ReportStrategyConfig)
                and self.report_period == other.report_period
                and self.report_strategy == other.report_strategy
                and self.aggregation_function == other.aggregation_function
                and self.ttl == other.ttl)

    def __str__(self):
        return f"ReportStrategyConfig(report_period={self.report_period}, report_strategy={self.report_strategy},\
            aggregation_function={self.aggregation_function}, ttl={self.ttl})"



from bacpypes3.appservice import ApplicationServiceAccessPoint as AppServiceAccessPoint, ServiceAccessPoint
from bacpypes3.apdu import (
    APCISequence,
    APDU,
    AbortPDU,
    ComplexAckPDU,
    ConfirmedRequestPDU,
    ErrorPDU,
    RejectPDU,
    SimpleAckPDU,
    UnconfirmedRequestPDU,
)


class ApplicationServiceAccessPoint(AppServiceAccessPoint):
    async def sap_response(self, apdu: APDU) -> None:
        if isinstance(apdu, SimpleAckPDU):
            xpdu = apdu

        elif isinstance(
            apdu, (UnconfirmedRequestPDU, ConfirmedRequestPDU, ComplexAckPDU, ErrorPDU)
        ):
            try:
                xpdu = APCISequence.decode(apdu)
            except Exception:
                if isinstance(apdu, UnconfirmedRequestPDU):
                    return

                if isinstance(apdu, ConfirmedRequestPDU):
                    return

                xpdu = apdu
        elif isinstance(apdu, (RejectPDU, AbortPDU)):
            xpdu = apdu
        else:
            raise RuntimeError(f"invalid APDU (10): {type(apdu)}")

        await ServiceAccessPoint.sap_response(self, xpdu)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from enum import Enum
from re import findall, sub

from thingsboard_gateway.gateway.constants import (
    DATA_PARAMETER,
    DEVICE_SECTION_PARAMETER,
    RPC_ID_PARAMETER,
    RPC_PARAMS_PARAMETER,
    RPC_METHOD_PARAMETER,
    RPC_CONNECTOR_ARGUMENTS_PARAMETER
)

RPC_SET_SPLIT_PATTERNS = ["; ", ";=", "=", " "]


class OpcUaRpcType(Enum):
    CONNECTOR = 'CONNECTOR'
    DEVICE = 'DEVICE'
    RESERVED = 'RESERVED'


class OpcUaRpcRequest:
    def __init__(self, content: dict) -> None:
        if content.get('data') is None:
            content['data'] = {'params': content['params'], 'method': content['method'], 'id': content['id']}
        self.rpc_method = content[DATA_PARAMETER].get(RPC_METHOD_PARAMETER)
        self.rpc_type = self._get_rpc_type(content)
        self.timeout = content.get('timeout', 5)
        self.params = content[DATA_PARAMETER].get(RPC_PARAMS_PARAMETER)
        self._arguments = None
        self._received_identifier = None
        self.id = content[DATA_PARAMETER].get(RPC_ID_PARAMETER)
        self.device_name = content.get(DEVICE_SECTION_PARAMETER)
        self._fill_rpc_request(content)

    @property
    def arguments(self):
        return self._arguments

    @arguments.setter
    def arguments(self, value):
        self._arguments = value

    @property
    def received_identifier(self):
        return self._received_identifier

    @received_identifier.setter
    def received_identifier(self, value):
        self._received_identifier = value

    def _get_rpc_type(self, content: dict) -> OpcUaRpcType:
        if self._is_connector_rpc():
            return OpcUaRpcType.CONNECTOR

        elif self._is_reserved_rpc(content):
            return OpcUaRpcType.RESERVED

        elif content.get(DEVICE_SECTION_PARAMETER) and not self._is_reserved_rpc(content):
            return OpcUaRpcType.DEVICE

        raise ValueError('Unknown RPC type for content: %r' % content)

    def _is_connector_rpc(self) -> bool:
        try:
            (connector_type, *rpc_method_name) = self.rpc_method.split('_')
            if connector_type == "opcua":
                self.rpc_method = '_'.join(rpc_method_name)
                return True

        except (ValueError, AttributeError):
            return False

    @staticmethod
    def _is_reserved_rpc(content: dict) -> bool:
        rpc_method = content[DATA_PARAMETER][RPC_METHOD_PARAMETER].lower()
        if rpc_method == 'get' or rpc_method == 'set' and content.get(DEVICE_SECTION_PARAMETER):
            return True

        return False

    def _fill_rpc_request(self, content: dict):
        if self.rpc_type == OpcUaRpcType.CONNECTOR:
            self._fill_connector_rpc_request(content)
        elif self.rpc_type == OpcUaRpcType.DEVICE:
            self._fill_device_rpc_request(content)
        elif self.rpc_type == OpcUaRpcType.RESERVED:
            self._fill_reserved_rpc_request(content)

    def _fill_connector_rpc_request(self, content: dict):
        values_sections = content[DATA_PARAMETER][RPC_PARAMS_PARAMETER][RPC_CONNECTOR_ARGUMENTS_PARAMETER]
        values = [value["value"] for value in values_sections]
        self._arguments = values

    def _fill_reserved_rpc_request(self, content: dict):
        params = content.get(DATA_PARAMETER, {}).get(RPC_PARAMS_PARAMETER)
        if self.rpc_method == "get":
            self.params = self.params.rstrip(' ;\t\n\r')
            return
        if self.rpc_method != "set" or not isinstance(params, str):
            return
        ident = self.__find_identifier_request(params)

        if ident:
            value = params[len(ident):]
            delimiter = next((p for p in RPC_SET_SPLIT_PATTERNS if p in params), None)
            self.params = ident
            self.arguments = sub(r"[;\s]+$", "", value.split(delimiter)[-1]).strip(' ;')
            return

        delimiter = next((p for p in RPC_SET_SPLIT_PATTERNS if p in params), None)
        parts = (
            [p.strip() for p in params.split(delimiter) if p.strip()]
            if delimiter
            else [params.strip()]
        )

        if not delimiter or len(parts) != 2:
            return

        full_path, value = parts
        self.params = full_path
        self.arguments = sub(r"[;\s]+$", "", value)

    def _fill_device_rpc_request(self, content: dict):
        rpc_section = content[DATA_PARAMETER].get(RPC_METHOD_PARAMETER)
        self.rpc_method = rpc_section
        self.arguments = self.params

    @staticmethod
    def __find_identifier_request(path: str) -> list | None:
        identifier = findall(r"(ns=\d+;[isgb]=[^;\s}]+)(?=;|\s|$)", path)
        if identifier:
            return identifier[0]




#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from typing import List, Union

from thingsboard_gateway.gateway.constants import ATTRIBUTES_PARAMETER, TELEMETRY_PARAMETER, TIMESERIES_PARAMETER, \
    METADATA_PARAMETER
from thingsboard_gateway.gateway.entities.attributes import Attributes
from thingsboard_gateway.gateway.entities.datapoint_key import DatapointKey
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


def split_large_entries(entries: dict, first_item_max_data_size: int, max_data_size: int, ts=None, ts_size=None):
    split_chunks = []
    split_chunk_sizes = []
    current_chunk = {}
    current_size = 0
    ts_check = False

    for original_key, value in entries.items():
        entry_size = TBUtility.get_data_size({original_key.key: value}) + 1
        if ts is not None and not ts_check:
            entry_size += ts_size
            ts_check = True

        # Recalculate size after each addition and check if it exceeds the max size
        if current_size + entry_size >= (first_item_max_data_size if not split_chunks else max_data_size):
            # Append current chunk if it exceeds the max size
            if current_chunk:
                split_chunks.append(current_chunk)
                split_chunk_sizes.append(current_size)
                ts_check = False
            # Start a new chunk
            current_chunk = {original_key: value}  # New dict is created to avoid modifying the original dict
            current_size = entry_size
        else:
            # Add to current chunk
            current_chunk[original_key] = value
            current_size += entry_size

    # Add the last chunk if any
    if current_chunk:
        split_chunks.append(current_chunk)
        split_chunk_sizes.append(current_size)

    return zip(split_chunks, split_chunk_sizes)


class ConvertedData:
    __slots__ = ['device_name', 'device_type', 'telemetry', 'attributes',
                 'metadata', '_telemetry_datapoints_count', 'ts_index']

    def __init__(self, device_name, device_type='default', metadata=None):
        self.device_name = device_name
        self.device_type = device_type
        self.telemetry: List[TelemetryEntry] = []
        self.attributes: Attributes = Attributes()
        self._telemetry_datapoints_count = 0
        self.metadata = metadata or {}
        self.ts_index = {}

    def __eq__(self, other):
        if not isinstance(other, ConvertedData):
            return False
        return self.to_dict() == other.to_dict()

    def __str__(self):
        return f"ConvertedData(deviceName={self.device_name}, deviceType={self.device_type}, " \
               f"telemetry={self.telemetry}, attributes={self.attributes}, metadata={self.metadata})"

    def __repr__(self):
        return self.__str__()

    def to_dict(self, debug_enabled=False):
        result = {
            "deviceName": self.device_name,
            "deviceType": self.device_type,
            "telemetry": [telemetry_entry.to_dict() for telemetry_entry in self.telemetry],
            "attributes": self.attributes.to_dict()
        }
        if debug_enabled:
            result["metadata"] = self.metadata
        return result

    def __getitem__(self, item):
        if item == ATTRIBUTES_PARAMETER:
            return self.attributes
        elif item in {TELEMETRY_PARAMETER, TIMESERIES_PARAMETER}:
            return self.telemetry
        elif item == METADATA_PARAMETER:
            return self.metadata
        else:
            raise KeyError(f"{item} - Item not found!")

    def extend(self, other: 'ConvertedData'):
        if not isinstance(other, ConvertedData):
            raise ValueError("Can only extend with another ConvertedData object.")
        self.telemetry.extend(other.telemetry)
        self.attributes.update(other.attributes)
        self.metadata.update(other.metadata)
        self._telemetry_datapoints_count += other.telemetry_datapoints_count
        for telemetry_entry in self.telemetry:
            if telemetry_entry.ts in self.ts_index:
                index = self.ts_index[telemetry_entry.ts]
                existing_values = self.telemetry[index].values
                existing_values.update(telemetry_entry.values)
            else:
                self.ts_index[telemetry_entry.ts] = len(self.telemetry) - 1

    def add_to_telemetry(self, telemetry_entry: Union[dict, TelemetryEntry, List[TelemetryEntry], List[dict]]):
        if isinstance(telemetry_entry, list):
            for entry in telemetry_entry:
                self._add_single_telemetry_entry(entry)
        else:
            self._add_single_telemetry_entry(telemetry_entry)

    def _add_single_telemetry_entry(self, telemetry_entry: Union[dict, TelemetryEntry]):
        if isinstance(telemetry_entry, dict):
            telemetry_entry = TelemetryEntry(telemetry_entry, telemetry_entry.get("ts"))

        if telemetry_entry.ts in self.ts_index:
            index = self.ts_index[telemetry_entry.ts]
            existing_values = self.telemetry[index].values
            old_values_len = len(existing_values)

            self.telemetry[index].values.update(telemetry_entry.values)
            self._telemetry_datapoints_count -= old_values_len
            self._telemetry_datapoints_count += len(self.telemetry[index].values)
        else:
            self.telemetry.append(telemetry_entry)
            self._telemetry_datapoints_count += len(telemetry_entry.values)
            self.ts_index[telemetry_entry.ts] = len(self.telemetry) - 1

    def add_to_attributes(self, key_or_entry: Union[dict, str, List[dict], DatapointKey], value=None):
        if isinstance(key_or_entry, list):
            for entry in key_or_entry:
                if not isinstance(entry, dict):
                    raise ValueError("Batch attribute processing requires a list of dictionaries.")
                self.attributes.update(entry)
        else:
            if isinstance(key_or_entry, str) or isinstance(key_or_entry, DatapointKey):
                if value is None:
                    raise ValueError("Invalid arguments for add_attribute")
                key_or_entry = {key_or_entry: value}

            self.attributes.update(key_or_entry)

    def add_to_metadata(self, key_value_entry: dict):
        self.metadata.update(key_value_entry)

    def get_size(self):
        return TBUtility.get_data_size(self.to_dict())

    @property
    def telemetry_datapoints_count(self):
        return self._telemetry_datapoints_count

    @property
    def attributes_datapoints_count(self):
        return len(self.attributes)

    # Methods for getting data
    def convert_to_objects_with_maximal_size(self, max_data_size) -> List['ConvertedData']:
        general_info_bytes_size = TBUtility.get_data_size({
            "deviceName": self.device_name,
            "deviceType": self.device_type,
            "metadata": self.metadata,
            "telemetry": [],
            "attributes": {}
        })

        if general_info_bytes_size > max_data_size:
            raise ValueError("Maximal data size is too small even for general info, please adjust maxPayloadSize")

        available_data_size = max_data_size - general_info_bytes_size

        converted_objects = []
        current_data = ConvertedData(self.device_name, self.device_type, self.metadata)
        current_data_size = general_info_bytes_size

        attributes_dict = self.attributes.to_dict()
        if attributes_dict:
            attributes_bytes_size = TBUtility.get_data_size(attributes_dict)
            if current_data_size + attributes_bytes_size <= max_data_size:
                current_data.add_to_attributes(attributes_dict)
                current_data_size += attributes_bytes_size
            else:
                split_attributes_and_sizes = split_large_entries(attributes_dict,
                                                                 max_data_size - current_data_size,
                                                                 available_data_size)
                for data_chunk, chunk_size in split_attributes_and_sizes:
                    if current_data_size + chunk_size >= max_data_size:
                        converted_objects.append(current_data)
                        current_data = ConvertedData(self.device_name, self.device_type, self.metadata)
                        current_data_size = general_info_bytes_size + chunk_size
                    current_data.add_to_attributes(data_chunk)
                    current_data_size += chunk_size

        for telemetry_entry in self.telemetry:
            telemetry_values = telemetry_entry.values
            ts_data_size = TBUtility.get_data_size({"ts": telemetry_entry.ts}) + 1

            telemetry_obj_size = TBUtility.get_data_size(
                {datapoint_key.key: value for datapoint_key, value in telemetry_values.items()}) + ts_data_size

            if telemetry_obj_size <= max_data_size - current_data_size:
                current_data.add_to_telemetry(telemetry_entry)
                current_data_size += telemetry_obj_size
            else:
                split_telemetry_and_sizes = split_large_entries(telemetry_values,
                                                                max_data_size - current_data_size,
                                                                available_data_size,
                                                                telemetry_entry.ts,
                                                                ts_data_size)
                for telemetry_chunk, chunk_size in split_telemetry_and_sizes:

                    if current_data_size + chunk_size > max_data_size:
                        converted_objects.append(current_data)
                        current_data = ConvertedData(self.device_name, self.device_type, self.metadata)
                        current_data_size = general_info_bytes_size
                    current_data_size += chunk_size
                    if telemetry_entry.ts in current_data.ts_index:
                        current_data_size += ts_data_size
                    current_data.add_to_telemetry(TelemetryEntry(telemetry_chunk, telemetry_entry.ts))

        if current_data_size > general_info_bytes_size:
            converted_objects.append(current_data)

        return converted_objects



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from enum import Enum

from simplejson import dumps
from time import time


# Service constants

DEV_MODE_PARAMETER_NAME = 'TB_GW_DEV_MODE'
TB_GW_DEV_DEBUG_SERVER_PORT = 5678

STATISTIC_MESSAGE_RECEIVED_PARAMETER = "MessagesReceived"
STATISTIC_MESSAGE_SENT_PARAMETER = "MessagesSent"

CONNECTOR_PARAMETER = "connector"
CONVERTER_PARAMETER = "converter"
UPLINK_PREFIX = "uplink_"
DOWNLINK_PREFIX = "downlink_"

CONFIG_SECTION_PARAMETER = "config"
CONFIG_VERSION_PARAMETER = "configVersion"
CONFIG_SERVER_SECTION_PARAMETER = "server"
CONFIG_DEVICES_SECTION_PARAMETER = "devices"

CONNECTED_DEVICES_FILENAME = "connected_devices.json"
PERSISTENT_GRPC_CONNECTORS_KEY_FILENAME = "persistent_keys.json"

RENAMING_PARAMETER = "renaming"
DISCONNECTED_PARAMETER = "disconnected"
CONNECTOR_ID_PARAMETER = "connectorId"
CONNECTOR_NAME_PARAMETER = "connectorName"

SECURITY_VAR = ('type', 'accessToken', 'caCert', 'privateKey', 'cert', 'clientId', 'username', 'password')

# Data parameter constants

DEVICE_SECTION_PARAMETER = "device"

DEVICE_NAME_PARAMETER = "deviceName"
DEVICE_TYPE_PARAMETER = "deviceType"

ATTRIBUTES_PARAMETER = "attributes"
ON_ATTRIBUTE_UPDATE_DEFAULT_TIMEOUT = 5
TELEMETRY_PARAMETER = "telemetry"
TIMESERIES_PARAMETER = "timeseries"
TELEMETRY_TIMESTAMP_PARAMETER = "ts"
TELEMETRY_VALUES_PARAMETER = "values"
METADATA_PARAMETER = "metadata"

# Messages metadata constants
RECEIVED_TS_PARAMETER = "receivedTs"
CONVERTED_TS_PARAMETER = "convertedTs"
SEND_TO_STORAGE_TS_PARAMETER = "sendToStorageTs"
DATA_RETRIEVING_STARTED = "dataRetrieveStartedTs"

# Size of metadata that will be added to messages in debug mode
# Connector name length should be added to the size of the metadata
DEBUG_METADATA_TEMPLATE_SIZE = len(dumps({
    "metadata": {
        RECEIVED_TS_PARAMETER: int(time() * 1000),
        CONNECTOR_PARAMETER: "",
        "publishedTs": int(time() * 1000)}}, separators=(',', ':'), skipkeys=True).encode("utf-8"))

# Report strategy parameters
REPORT_STRATEGY_PARAMETER = "reportStrategy"
REPORT_PERIOD_PARAMETER = "reportPeriod"
TYPE_PARAMETER = "type"
AGGREGATION_FUNCTION_PARAMETER = "aggregationFunction"
TTL_PARAMETER = "ttl"


class ReportStrategy(Enum):
    ON_REPORT_PERIOD = "ON_REPORT_PERIOD"
    ON_CHANGE = "ON_CHANGE"
    ON_CHANGE_OR_REPORT_PERIOD = "ON_CHANGE_OR_REPORT_PERIOD"
    ON_RECEIVED = "ON_RECEIVED"
    DISABLED = "DISABLED"

    @classmethod
    def from_string(cls, value: str):
        for strategy in cls:
            if strategy.value.lower() == value.lower():
                return strategy
        raise ValueError("Invalid report strategy value: %r" % value)


DEFAULT_REPORT_STRATEGY_CONFIG = {
    TYPE_PARAMETER: ReportStrategy.ON_RECEIVED.value,
    REPORT_PERIOD_PARAMETER: 10000,
    TTL_PARAMETER: 86400
}

STRATEGIES_WITH_REPORT_PERIOD = (ReportStrategy.ON_REPORT_PERIOD, ReportStrategy.ON_CHANGE_OR_REPORT_PERIOD)

# RPC parameter constants

RPC_ID_PARAMETER = "id"
RPC_METHOD_PARAMETER = "method"
RPC_PARAMS_PARAMETER = "params"
DATA_PARAMETER = "data"
RPC_DEFAULT_TIMEOUT=5
RPC_CONNECTOR_ARGUMENTS_PARAMETER = "arguments"
RPC_VALUE_PARAMETER = "value"

# Attribute constants
ATTRIBUTES_FOR_REQUEST = ["RemoteLoggingLevel", "general_configuration", "storage_configuration",
                          "grpc_configuration", "logs_configuration", "active_connectors"]


DEFAULT_CONNECTORS = {
    "mqtt": "MqttConnector",
    "modbus": "AsyncModbusConnector",
    "opcua": "OpcUaConnector",
    "ble": "BLEConnector",
    "request": "RequestConnector",
    "can": "CanConnector",
    "bacnet": "AsyncBACnetConnector",
    "odbc": "OdbcConnector",
    "rest": "RESTConnector",
    "snmp": "SNMPConnector",
    "ftp": "FTPConnector",
    "socket": "SocketConnector",
    "xmpp": "XMPPConnector",
    "ocpp": "OcppConnector",
    "knx": "KNXConnector",
}

DEFAULT_STATISTIC = {
    'enable': True,
    'statsSendPeriodInSeconds': 60,
    'customStatsSendPeriodInSeconds': 3600
}

DEFAULT_DEVICE_FILTER = {
    'enable': False
}

CUSTOM_RPC_DIR = "/etc/thingsboard-gateway/rpc"

# Provisioning constants

PROVISIONED_CREDENTIALS_FILENAME = ".credentials"
